"""
Tests for scale components module.
"""

from __future__ import annotations

from math import exp

import pytest

from scitbx import sparse

from dials.algorithms.scaling.model.components.scale_components import (
    ScaleComponentBase,
    SHScaleComponent,
    SingleBScaleFactor,
    SingleScaleFactor,
)
from dials.algorithms.scaling.model.components.smooth_scale_components import (
    SmoothBScaleComponent1D,
    SmoothMixin,
    SmoothScaleComponent1D,
    SmoothScaleComponent2D,
    SmoothScaleComponent3D,
)
from dials.array_family import flex


def test_ScaleComponentBase():
    """Test for the ScaleComponentBase class."""

    # Test initialisation with no parameter esds.
    base_SF = ScaleComponentBase(flex.double([1.0] * 3))
    assert base_SF.n_params == 3
    assert base_SF.parameter_esds is None
    assert list(base_SF.parameters) == [1.0, 1.0, 1.0]

    # Test updating of parameters
    base_SF.parameters = flex.double([2.0, 2.0, 2.0])
    assert list(base_SF.parameters) == [2.0, 2.0, 2.0]
    with pytest.raises(AssertionError):
        # Try to change the number of parameters - should fail
        base_SF.parameters = flex.double([2.0, 2.0, 2.0, 2.0])

    # Test setting of var_cov matrix.
    assert base_SF.var_cov_matrix is None
    base_SF.var_cov_matrix = [1.0, 0.0, 0.0]
    assert base_SF.var_cov_matrix == [1.0, 0.0, 0.0]

    assert base_SF.calculate_restraints() is None
    assert base_SF.calculate_jacobian_restraints() is None

    base_SF = ScaleComponentBase(flex.double(3, 1.0), flex.double(3, 0.1))
    assert list(base_SF.parameter_esds) == [0.1] * 3


def test_SingleScaleFactor():
    """Test for SingleScaleFactor class."""
    KSF = SingleScaleFactor(flex.double([2.0]))
    assert KSF.n_params == 1
    assert list(KSF.parameters) == [2.0]
    rt = flex.reflection_table()
    rt["d"] = flex.double([1.0, 1.0])
    rt["id"] = flex.int([0, 0])
    KSF.data = {"id": rt["id"]}
    KSF.update_reflection_data()
    assert KSF.n_refl == [2]
    s, d = KSF.calculate_scales_and_derivatives()
    assert list(s) == [2.0, 2.0]
    assert d[0, 0] == 1
    assert d[1, 0] == 1
    s, d = KSF.calculate_scales_and_derivatives()
    KSF.update_reflection_data(flex.bool([True, False]))  # Test selection.
    assert KSF.n_refl[0] == 1


def test_SingleBScaleFactor():
    """Test forSingleBScaleFactor class."""
    BSF = SingleBScaleFactor(flex.double([0.0]))
    assert BSF.n_params == 1
    assert list(BSF.parameters) == [0.0]
    rt = flex.reflection_table()
    rt["d"] = flex.double([1.0, 1.0])
    # rt["id"] = flex.int([0, 0])
    BSF.data = {"d": rt["d"]}
    BSF.update_reflection_data()
    assert BSF.n_refl == [2]
    assert list(BSF.d_values[0]) == [1.0, 1.0]
    s, d = BSF.calculate_scales_and_derivatives()
    assert list(s) == [1.0, 1.0]
    assert d[0, 0] == 0.5
    assert d[1, 0] == 0.5
    s, d = BSF.calculate_scales_and_derivatives()
    BSF.update_reflection_data(flex.bool([True, False]))  # Test selection.
    assert BSF.n_refl[0] == 1


def test_SHScalefactor():
    """Test the spherical harmonic absorption component."""
    initial_param = 0.1
    initial_val = 0.2

    SF = SHScaleComponent(flex.double([initial_param] * 3))
    assert SF.n_params == 3
    assert list(SF.parameters) == [initial_param] * 3

    # Test functionality just by setting sph_harm_table directly and calling
    # update_reflection_data to initialise the harmonic values.
    harmonic_values = sparse.matrix(3, 1)
    harmonic_values[0, 0] = initial_val
    harmonic_values[1, 0] = initial_val
    harmonic_values[2, 0] = initial_val
    SF.data = {"sph_harm_table": harmonic_values}
    SF.update_reflection_data()
    print(SF.harmonic_values)
    assert SF.harmonic_values[0][0, 0] == initial_val
    assert SF.harmonic_values[0][0, 1] == initial_val
    assert SF.harmonic_values[0][0, 2] == initial_val
    s, d = SF.calculate_scales_and_derivatives()
    assert list(s) == [1.0 + (3.0 * initial_val * initial_param)]
    assert d[0, 0] == initial_val
    assert d[0, 1] == initial_val
    assert d[0, 2] == initial_val
    s, d = SF.calculate_scales_and_derivatives()

    # Test functionality of passing in a selection
    harmonic_values = sparse.matrix(3, 2)
    harmonic_values[0, 0] = initial_val
    harmonic_values[0, 1] = initial_val
    harmonic_values[2, 0] = initial_val
    SF.data = {"sph_harm_table": harmonic_values}
    SF.update_reflection_data(flex.bool([False, True]))
    assert SF.harmonic_values[0].n_rows == 1
    assert SF.harmonic_values[0].n_cols == 3
    assert SF.n_refl[0] == 1

    # Test setting of restraints and that restraints are calculated.
    # Not testing actual calculation as may want to change the form.
    SF.parameter_restraints = flex.double([0.1, 0.2, 0.3])
    assert SF.parameter_restraints == flex.double([0.1, 0.2, 0.3])
    restraints = SF.calculate_restraints()
    assert restraints[0] is not None
    assert restraints[1] is not None
    jacobian_restraints = SF.calculate_jacobian_restraints()
    assert jacobian_restraints[0] is not None
    assert jacobian_restraints[1] is not None


def test_SmoothMixin():
    """Simple test for the Smooth Mixin class."""
    Smooth_mixin_class = SmoothMixin()
    assert hasattr(Smooth_mixin_class, "smoother")
    assert Smooth_mixin_class.nparam_to_val(2) == 1
    assert Smooth_mixin_class.nparam_to_val(3) == 2
    assert Smooth_mixin_class.nparam_to_val(5) == 3
    assert Smooth_mixin_class.nparam_to_val(6) == 4


def test_SmoothScaleFactor1D():
    """Test for the gaussian smoothed 1D scalefactor class."""
    SF = SmoothScaleComponent1D(flex.double(5, 1.1))
    assert SF.n_params == 5
    assert list(SF.parameters) == [1.1, 1.1, 1.1, 1.1, 1.1]
    norm_rot = flex.double([0.5, 1.0, 2.5, 0.0])
    SF.data = {"x": norm_rot}
    SF.update_reflection_data()
    assert list(SF.normalised_values[0]) == [0.5, 1.0, 2.5, 0.0]
    SF.smoother.set_smoothing(4, 1.0)
    assert list(SF.smoother.positions()) == [-0.5, 0.5, 1.5, 2.5, 3.5]
    s, d = SF.calculate_scales_and_derivatives()
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)
    assert list(s) == pytest.approx([1.1, 1.1, 1.1, 1.1])
    assert d[0, 0] / d[0, 1] == pytest.approx(exp(-1.0) / exp(0.0))
    sumexp = exp(-1.0 / 1.0) + exp(-0.0 / 1.0) + exp(-1.0 / 1.0)  # only averages 3 when
    # normalised position is exactly on a smoother position.
    assert d[0, 1] == pytest.approx(exp(0.0) / sumexp)
    T = d.transpose()
    assert sum(list(T[:, 0].as_dense_vector())) == 1.0  # should always be 1.0
    assert sum(list(T[:, 1].as_dense_vector())) == 1.0
    assert sum(list(T[:, 2].as_dense_vector())) == 1.0
    assert d[1, 1] == d[1, 2]
    assert d[1, 0] == d[1, 3]
    s, d = SF.calculate_scales_and_derivatives()

    # Test that if one or none in block, then doesn't fail but returns sensible value
    SF._normalised_values = [flex.double([0.5])]
    SF._n_refl = [1]
    assert list(SF.normalised_values[0]) == [0.5]
    s, d = SF.calculate_scales_and_derivatives()
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)
    SF._normalised_values = [flex.double()]
    SF._n_refl = [0]
    assert list(SF.normalised_values[0]) == []
    s, d = SF.calculate_scales_and_derivatives()
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)


def test_SmoothBScaleFactor1D():
    "test for a gaussian smoothed 1D scalefactor object"
    SF = SmoothBScaleComponent1D(flex.double(5, 0.0))
    assert SF.n_params == 5
    assert list(SF.parameters) == [0.0] * 5
    norm_rot = flex.double([0.5, 1.0, 2.5, 0.0])
    d = flex.double([1.0, 1.0, 1.0, 1.0])
    SF.data = {"x": norm_rot, "d": d}
    SF.update_reflection_data()
    assert list(SF.normalised_values[0]) == [0.5, 1.0, 2.5, 0.0]
    assert list(SF.d_values[0]) == [1.0, 1.0, 1.0, 1.0]
    SF.smoother.set_smoothing(4, 1.0)
    s, d = SF.calculate_scales_and_derivatives()
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)
    assert list(s) == pytest.approx([1.0, 1.0, 1.0, 1.0])
    assert d[0, 0] / d[0, 1] == pytest.approx(
        exp(-1.0) / exp(0.0)
    )  # derivative ratio of two adjacent params (at +-0.5)
    sumexp = exp(-1.0 / 1.0) + exp(-0.0 / 1.0) + exp(-1.0 / 1.0)
    assert d[0, 1] == pytest.approx((exp(0.0) / sumexp) * s[1] / 2.0)
    T = d.transpose()
    assert sum(list(T[:, 0].as_dense_vector())) == 0.5  # value depends on d
    assert sum(list(T[:, 1].as_dense_vector())) == 0.5
    assert sum(list(T[:, 2].as_dense_vector())) == 0.5
    assert d[1, 1] == d[1, 2]
    assert d[1, 0] == d[1, 3]
    s, d = SF.calculate_scales_and_derivatives()

    SF._normalised_values = [flex.double([0.5])]
    SF._n_refl = [1]
    SF._d_values = [flex.double([1.0])]
    assert list(SF.normalised_values[0]) == [0.5]
    s, d = SF.calculate_scales_and_derivatives()
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)


def test_SmoothScaleFactor2D():
    """Test the 2D smooth scale factor class."""
    with pytest.raises(AssertionError):  # Test incorrect shape initialisation
        SF = SmoothScaleComponent2D(flex.double(30, 1.1), shape=(5, 5))
    SF = SmoothScaleComponent2D(flex.double(30, 1.1), shape=(6, 5))
    assert SF.n_x_params == 6
    assert SF.n_y_params == 5
    assert SF.n_params == 30

    assert list(SF.parameters) == [1.1] * 30
    norm_rot = flex.double(30, 0.5)
    norm_time = flex.double(30, 0.5)
    norm_rot[0] = 0.0
    norm_time[0] = 0.0
    norm_rot[29] = 3.99
    norm_time[29] = 2.99
    SF.data = {"x": norm_rot, "y": norm_time}
    SF.update_reflection_data()
    # assert list(SF.normalised_x_values) == list(flex.double(
    #  [0.5, 0.5, 0.5, 0.0, 0.0, 0.0]))
    # assert list(SF.normalised_y_values) == list(flex.double(
    #  [0.0, 0.0, 0.0, 0.5, 0.5, 0.5]))
    SF.smoother.set_smoothing(4, 1.0)  # will average 3 in x,y dims.
    assert list(SF.smoother.x_positions()) == [-0.5, 0.5, 1.5, 2.5, 3.5, 4.5]
    assert list(SF.smoother.y_positions()) == [-0.5, 0.5, 1.5, 2.5, 3.5]
    s, d = SF.calculate_scales_and_derivatives()
    assert list(s) == pytest.approx([1.1] * 30)
    sumexp = exp(0.0) + (4.0 * exp(-1.0 / 1.0)) + (4.0 * exp(-2.0 / 1.0))
    assert d[1, 7] == pytest.approx(exp(-0.0) / sumexp)

    # Test again with a small number of params to check different behaviour.
    SF = SmoothScaleComponent2D(flex.double(6, 1.1), shape=(3, 2))
    _ = flex.reflection_table()
    norm_rot = flex.double(6, 0.5)
    norm_time = flex.double(6, 0.5)
    norm_rot[0] = 0.0
    norm_time[0] = 0.0
    norm_rot[5] = 1.99
    norm_time[5] = 0.99
    SF.data = {"x": norm_rot, "y": norm_time}
    SF.update_reflection_data()
    SF.smoother.set_smoothing(4, 1.0)  # will average 3,2 in x,y dims.
    assert list(SF.smoother.x_positions()) == [0.0, 1.0, 2.0]
    assert list(SF.smoother.y_positions()) == [0.0, 1.0]
    s, d = SF.calculate_scales_and_derivatives()
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)
    sumexp = (4.0 * exp(-0.5 / 1.0)) + (2.0 * exp(-2.5 / 1.0))
    assert d[1, 1] == pytest.approx(exp(-0.5) / sumexp)

    # Test that if one or none in block, then doesn't fail but returns sensible value
    SF._normalised_x_values = [flex.double([0.5])]
    SF._normalised_y_values = [flex.double([0.5])]
    SF._n_refl = [1]
    assert list(SF.normalised_x_values[0]) == [0.5]
    assert list(SF.normalised_y_values[0]) == [0.5]
    s, d = SF.calculate_scales_and_derivatives()
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)
    assert list(s) == pytest.approx([1.1])
    SF._normalised_x_values = [flex.double()]
    SF._normalised_y_values = [flex.double()]
    SF._n_refl = [0]
    assert list(SF.normalised_x_values[0]) == []
    assert list(SF.normalised_y_values[0]) == []
    s, d = SF.calculate_scales_and_derivatives()
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)
    assert list(s) == []
    assert d == sparse.matrix(0, 0)


def test_SmoothScaleFactor3D():
    """Test the 2D smooth scale factor class."""
    with pytest.raises(AssertionError):  # Test incorrect shape initialisation
        SF = SmoothScaleComponent3D(flex.double(150, 1.1), shape=(5, 5, 5))
    SF = SmoothScaleComponent3D(flex.double(150, 1.1), shape=(6, 5, 5))
    assert SF.n_x_params == 6
    assert SF.n_y_params == 5
    assert SF.n_z_params == 5
    assert SF.n_params == 150

    assert list(SF.parameters) == [1.1] * 150
    norm_rot = flex.double(150, 0.5)
    norm_time = flex.double(150, 0.5)
    norm_z = flex.double(150, 0.5)
    norm_rot[0] = 0.0
    norm_time[0] = 0.0
    norm_z[0] = 0.0
    norm_rot[149] = 3.99
    norm_time[149] = 2.99
    norm_z[149] = 2.99
    SF.data = {"x": norm_rot, "y": norm_time, "z": norm_z}
    SF.update_reflection_data()
    SF.smoother.set_smoothing(3, 1.0)  # will average 3 in x,y,z dims for test.
    assert list(SF.smoother.x_positions()) == [-0.5, 0.5, 1.5, 2.5, 3.5, 4.5]
    assert list(SF.smoother.y_positions()) == [-0.5, 0.5, 1.5, 2.5, 3.5]
    assert list(SF.smoother.z_positions()) == [-0.5, 0.5, 1.5, 2.5, 3.5]
    s, d = SF.calculate_scales_and_derivatives()
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)
    assert list(s) == pytest.approx([1.1] * 150)
    sumexp = (
        exp(-0.0)
        + (6.0 * exp(-1.0 / 1.0))
        + (8.0 * exp(-3.0 / 1.0))
        + (12.0 * exp(-2.0 / 1.0))
    )
    assert d[1, 7] == pytest.approx(exp(-1.0) / sumexp)  # Just check one

    # Test that if one or none in block, then doesn't fail but returns sensible value
    SF._normalised_x_values = [flex.double([0.5])]
    SF._normalised_y_values = [flex.double([0.5])]
    SF._normalised_z_values = [flex.double([0.5])]
    SF._n_refl = [1]
    assert list(SF.normalised_x_values[0]) == [0.5]
    assert list(SF.normalised_y_values[0]) == [0.5]
    assert list(SF.normalised_z_values[0]) == [0.5]
    s, d = SF.calculate_scales_and_derivatives()
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)
    assert list(s) == pytest.approx([1.1])
    SF._normalised_x_values = [flex.double()]
    SF._normalised_y_values = [flex.double()]
    SF._normalised_z_values = [flex.double()]
    SF._n_refl = [0]
    assert list(SF.normalised_x_values[0]) == []
    assert list(SF.normalised_y_values[0]) == []
    assert list(SF.normalised_z_values[0]) == []
    s, d = SF.calculate_scales_and_derivatives()
    assert list(s) == []
    s2 = SF.calculate_scales()
    assert list(s) == list(s2)
    assert d == sparse.matrix(0, 0)



from __future__ import annotations

import pytest

import iotbx.mtz
from cctbx import sgtbx
from cctbx.array_family import flex
from dxtbx.model import Experiment, ExperimentList, Scan

import dials.pychef


def test_observations():
    miller_indices = flex.miller_index(
        ((1, 2, 3), (-1, 2, 3), (1, -2, 3), (4, 5, 6), (4, 5, -6))
    )
    sg = sgtbx.space_group_info(symbol="I222").group()
    anomalous_flag = True
    observations = dials.pychef.Observations(miller_indices, sg, anomalous_flag)
    groups = observations.observation_groups()
    assert list(groups[(1, 2, 3)].iplus()) == [0]
    assert list(groups[(1, 2, 3)].iminus()) == [1, 2]


def test_accumulators(dials_data):
    f = dials_data("pychef", pathlib=True) / "insulin_dials_scaled_unmerged.mtz"
    mtz_object = iotbx.mtz.object(file_name=str(f))
    arrays = mtz_object.as_miller_arrays(merge_equivalents=False)
    for ma in arrays:
        if ma.info().labels == ["BATCH"]:
            batches = ma
        elif ma.info().labels == ["I", "SIGI"]:
            intensities = ma
        elif ma.info().labels == ["I(+)", "SIGI(+)", "I(-)", "SIGI(-)"]:
            intensities = ma
    assert intensities is not None
    assert batches is not None

    anomalous_flag = True
    if anomalous_flag:
        intensities = intensities.as_anomalous_array()

    stats = dials.pychef.Statistics(intensities, batches.data())

    # test completeness
    assert stats.iplus_comp_overall.size() == 46
    assert stats.iplus_comp_overall[45] == pytest.approx(0.9428352196431997)
    assert stats.iminus_comp_overall[44] == pytest.approx(0.07769038941108766)
    assert stats.ieither_comp_overall[1] == pytest.approx(0.03721465566852101)
    assert stats.iboth_comp_overall[-1] == pytest.approx(0.0779781315940917)

    # test rcp,scp
    print(list(stats.rcp))
    print(list(stats.scp))
    assert stats.rcp[45] == pytest.approx(0.04844584637191411)
    assert stats.scp[45] == pytest.approx(0.9201295853457298)

    # test Rd
    print(list(stats.rd))
    assert stats.rd[0] == pytest.approx(0.05234416616316846)


def test_interpret_images_to_doses_options():
    """Test handling of command line options for experiments input."""
    experiments = ExperimentList()
    experiments.append(Experiment(scan=Scan(image_range=(1, 10), oscillation=(0, 1.0))))
    experiments.append(Experiment(scan=Scan(image_range=(1, 20), oscillation=(0, 1.0))))
    experiments.append(Experiment(scan=Scan(image_range=(1, 10), oscillation=(0, 1.0))))

    # Default
    starting_doses, dpi = dials.pychef.interpret_images_to_doses_options(
        experiments, dose_per_image=[1]
    )
    assert starting_doses == [0, 0, 0]
    assert dpi == [1.0, 1.0, 1.0]

    # Multi-sweep measurements on same crystal
    starting_doses, dpi = dials.pychef.interpret_images_to_doses_options(
        experiments, dose_per_image=[1], shared_crystal=True
    )
    assert starting_doses == [0, 10, 30]
    assert dpi == [1.0, 1.0, 1.0]

    # Specify starting doses
    starting_doses, dpi = dials.pychef.interpret_images_to_doses_options(
        experiments, dose_per_image=[1], starting_doses=[0, 20, 0], shared_crystal=False
    )
    assert starting_doses == [0, 20, 0]
    assert dpi == [1.0, 1.0, 1.0]

    # Specify doses per image and shared crystal
    starting_doses, dpi = dials.pychef.interpret_images_to_doses_options(
        experiments, dose_per_image=[1.0, 2.0, 1.0], shared_crystal=True
    )
    assert starting_doses == [0, 10, 50]
    assert dpi == [1.0, 2.0, 1.0]

    # Test error is raised if bad input values for starting doses or dose per image.
    with pytest.raises(ValueError):
        _, __ = dials.pychef.interpret_images_to_doses_options(
            experiments, dose_per_image=[1.0], starting_doses=[0, 1]
        )
    with pytest.raises(ValueError):
        _, __ = dials.pychef.interpret_images_to_doses_options(
            experiments, dose_per_image=[1.0, 2.0]
        )



from __future__ import annotations

import copy
import itertools
import json
import math

import numpy as np

from cctbx import uctbx

import dials.util.log
from dials.algorithms.scaling.model.model import plot_scaling_models
from dials.algorithms.scaling.observers import make_merging_stats_plots
from dials.algorithms.scaling.scaling_library import (
    merging_stats_from_scaled_array,
    scaled_data_as_miller_array,
)
from dials.algorithms.scaling.scaling_utilities import DialsMergingStatisticsError
from dials.array_family import flex
from dials.report.plots import i_over_sig_i_vs_i_plot
from dials.util import show_mail_handle_errors
from dials.util.command_line import Command

RAD2DEG = 180 / math.pi

help_message = """

Generates a html report given the output of various DIALS programs
(observations.refl and/or models.expt).

Examples::

  dials.report strong.refl

  dials.report indexed.refl

  dials.report refined.refl

  dials.report integrated.refl

  dials.report refined.expt

  dials.report integrated.refl integrated.expt
"""

import libtbx.phil

# Create the phil parameters
phil_scope = libtbx.phil.parse(
    """
  output {
    html = dials.report.html
      .type = path
      .help = "The name of the output html file"
    json = None
      .type = path
      .help = "The name of the optional json file containing the plot data"
    external_dependencies = *remote local embed
      .type = choice
      .help = "Whether to use remote external dependencies (files relocatable"
              "but requires an internet connection), local (does not require"
              "internet connection but files may not be relocatable) or embed"
              "all external dependencies (inflates the html file size)."
  }
  grid_size = Auto
    .type = ints(size=2)
  pixels_per_bin = 40
    .type = int(value_min=1)

  centroid_diff_max = None
    .help = "Magnitude in pixels of shifts mapped to the extreme colours"
            "in the heatmap plots centroid_diff_x and centroid_diff_y"
    .type = float
    .expert_level = 1

  orientation_decomposition
    .help = "Options determining how the orientation matrix"
            "decomposition is done. The axes about which to decompose"
            "the matrix into three rotations are chosen here, as well"
            "as whether the rotations are relative to the reference"
            "orientation, taken from the static crystal model"
  {
    e1 = 1. 0. 0.
      .type = floats(size = 3)

    e2 = 0. 1. 0.
      .type = floats(size = 3)

    e3 = 0. 0. 1.
      .type = floats(size = 3)

    relative_to_static_orientation = True
      .type = bool
  }
"""
)


def ensure_directory(path):
    """Make the directory if not already there."""
    import errno
    from os import makedirs

    try:
        makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def ensure_required(rlist, required):
    """Check which keys aren't present."""
    not_present = []
    for k in required:
        if k not in rlist:
            not_present.append(k)
    if len(not_present) != 0:
        print(" Skipping: following required fields not present:")
        for k in not_present:
            print(f"  {k}")
        return False
    return True


def color_repeats(n=1):
    """Set up a cycle through default Plotly colors, repeating each n times"""

    color_list = (
        ["#1f77b4"] * n  # muted blue
        + ["#ff7f0e"] * n  # safety orange
        + ["#2ca02c"] * n  # cooked asparagus green
        + ["#d62728"] * n  # brick red
        + ["#9467bd"] * n  # muted purple
        + ["#8c564b"] * n  # chestnut brown
        + ["#e377c2"] * n  # raspberry yogurt pink
        + ["#7f7f7f"] * n  # middle gray
        + ["#bcbd22"] * n  # curry yellow-green
        + ["#17becf"] * n  # blue-teal
    )
    return itertools.cycle(color_list)


class ScanVaryingCrystalAnalyser:
    """Analyse a scan-varying crystal."""

    def __init__(self, orientation_decomposition):
        # Decomposition axes
        self._e1 = orientation_decomposition.e1
        self._e2 = orientation_decomposition.e2
        self._e3 = orientation_decomposition.e3
        self._relative_to_static_orientation = (
            orientation_decomposition.relative_to_static_orientation
        )
        self._debug = False

    def __call__(self, experiments):
        """Analyse the strong spots."""
        # Check we have the required fields
        print("Analysing scan-varying crystal model")

        d = {}

        if experiments is not None and len(experiments):
            d.update(self.plot_cell(experiments))
            d.update(self.plot_orientation(experiments))
        return {"scan_varying": d}

    def plot_cell(self, experiments):
        """Analyse the scan-varying cell parameters."""

        # cell plot
        dat = []
        for iexp, exp in enumerate(experiments):

            crystal = exp.crystal
            scan = exp.scan

            if crystal is None:
                print("Ignoring absent crystal model")
                continue

            if crystal.num_scan_points == 0:
                print("Ignoring scan-static crystal")
                continue

            scan_pts = list(range(crystal.num_scan_points))
            cells = [crystal.get_unit_cell_at_scan_point(t) for t in scan_pts]
            cell_params = [e.parameters() for e in cells]
            a, b, c, aa, bb, cc = zip(*cell_params)
            aa = [round(i, ndigits=6) for i in aa]
            bb = [round(i, ndigits=6) for i in bb]
            cc = [round(i, ndigits=6) for i in cc]
            start, stop = scan.get_array_range()
            phi = [scan.get_angle_from_array_index(t) for t in range(start, stop + 1)]
            vol = [e.volume() for e in cells]
            cell_dat = {
                "phi": phi,
                "a": a,
                "b": b,
                "c": c,
                "alpha": aa,
                "beta": bb,
                "gamma": cc,
                "volume": vol,
            }
            if self._debug:
                print(f"Crystal in Experiment {iexp}")
                print("Phi\ta\tb\tc\talpha\tbeta\tgamma\tVolume")
                msg = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}"
                line_dat = zip(phi, a, b, c, aa, bb, cc, vol)
                for line in line_dat:
                    print(msg.format(*line))
            dat.append(cell_dat)

        d = {
            "scan_varying_cell": {
                "data": [],
                "layout": {
                    "title": "Scan-varying cell parameters",
                    "xaxis3": {
                        "domain": [0, 1],
                        "anchor": "y7",
                        "title": "rotation angle (°)",
                    },
                    "xaxis2": {"domain": [0.53, 1], "anchor": "y6"},
                    "xaxis": {"domain": [0, 0.47], "anchor": "y3"},
                    "yaxis7": {"domain": [0.0, 0.2], "anchor": "x3", "nticks": 5},
                    "yaxis6": {"domain": [0.3, 0.5], "anchor": "x2", "nticks": 5},
                    "yaxis5": {"domain": [0.55, 0.75], "anchor": "x2", "nticks": 5},
                    "yaxis4": {"domain": [0.8, 1], "anchor": "x2", "nticks": 5},
                    "yaxis3": {"domain": [0.3, 0.5], "anchor": "x", "nticks": 5},
                    "yaxis2": {"domain": [0.55, 0.75], "anchor": "x", "nticks": 5},
                    "yaxis": {"domain": [0.8, 1], "nticks": 5},
                },
                "help": """\
A plot of the smoothly-varying unit cell parameters determined by scan-varying
refinement. Small variations up to ~1% of the linear cell dimensions are not
uncommon. Especially in cases of radiation damage, an increase of cell volume
with respect to dose is often seen. Multi-turn scans typically show some
oscillatory behaviour. This is not expected to be physically real, but indicates
the refinement algorithm accounting for unmodelled features in the data.
""",
            }
        }

        if len(dat) == 1:
            colors = color_repeats()
        else:
            colors = color_repeats(7)

        for cell_dat in dat:
            d["scan_varying_cell"]["data"].extend(
                [
                    {
                        "x": cell_dat["phi"],
                        "y": cell_dat["a"],
                        "type": "scatter",
                        "name": "a (Å)",
                        "marker": {"color": next(colors)},
                    },
                    {
                        "x": cell_dat["phi"],
                        "y": cell_dat["b"],
                        "type": "scatter",
                        "name": "b (Å)",
                        "xaxis": "x",
                        "yaxis": "y2",
                        "marker": {"color": next(colors)},
                    },
                    {
                        "x": cell_dat["phi"],
                        "y": cell_dat["c"],
                        "type": "scatter",
                        "name": "c (Å)",
                        "xaxis": "x",
                        "yaxis": "y3",
                        "marker": {"color": next(colors)},
                    },
                    {
                        "x": cell_dat["phi"],
                        "y": cell_dat["alpha"],
                        "type": "scatter",
                        "name": "α (°)",
                        "xaxis": "x2",
                        "yaxis": "y4",
                        "marker": {"color": next(colors)},
                    },
                    {
                        "x": cell_dat["phi"],
                        "y": cell_dat["beta"],
                        "type": "scatter",
                        "name": "β (°)",
                        "xaxis": "x2",
                        "yaxis": "y5",
                        "marker": {"color": next(colors)},
                    },
                    {
                        "x": cell_dat["phi"],
                        "y": cell_dat["gamma"],
                        "type": "scatter",
                        "name": "γ (°)",
                        "xaxis": "x2",
                        "yaxis": "y6",
                        "marker": {"color": next(colors)},
                    },
                    {
                        "x": cell_dat["phi"],
                        "y": cell_dat["volume"],
                        "type": "scatter",
                        "name": "volume (Å³)",
                        "xaxis": "x3",
                        "yaxis": "y7",
                        "marker": {"color": next(colors)},
                    },
                ]
            )
        if not len(dat):
            return {}
        return d

    def plot_orientation(self, experiments):
        from scitbx import matrix

        from dials.algorithms.refinement.rotation_decomposition import (
            solve_r3_rotation_for_angles_given_axes,
        )

        # orientation plot
        dat = []
        for iexp, exp in enumerate(experiments):

            crystal = exp.crystal
            scan = exp.scan

            if crystal is None:
                print("Ignoring absent crystal model")
                continue

            if crystal.num_scan_points == 0:
                print("Ignoring scan-static crystal")
                continue

            scan_pts = list(range(crystal.num_scan_points))
            start, stop = scan.get_array_range()
            phi = [scan.get_angle_from_array_index(t) for t in range(start, stop + 1)]
            Umats = [matrix.sqr(crystal.get_U_at_scan_point(t)) for t in scan_pts]
            if self._relative_to_static_orientation:
                # factor out static U
                Uinv = matrix.sqr(crystal.get_U()).inverse()
                Umats = [U * Uinv for U in Umats]
            # NB e3 and e1 definitions for the crystal are swapped compared
            # with those used inside the solve_r3_rotation_for_angles_given_axes
            # method
            angles = [
                solve_r3_rotation_for_angles_given_axes(
                    U, self._e3, self._e2, self._e1, deg=True
                )
                for U in Umats
            ]
            phi3, phi2, phi1 = zip(*angles)
            angle_dat = {"phi": phi, "phi3": phi3, "phi2": phi2, "phi1": phi1}
            if self._debug:
                print(f"Crystal in Experiment {iexp}")
                print("Image\tphi3\tphi2\tphi1")
                msg = "{0}\t{1}\t{2}\t{3}"
                line_dat = zip(phi, phi3, phi2, phi1)
                for line in line_dat:
                    print(msg.format(*line))
            dat.append(angle_dat)

        d = {
            "scan_varying_orientation": {
                "data": [],
                "layout": {
                    "title": "Scan-varying orientation parameters",
                    "xaxis3": {
                        "domain": [0, 1],
                        "anchor": "y3",
                        "title": "rotation angle (°)",
                    },
                    "xaxis2": {"domain": [0, 1], "anchor": "y3"},
                    "xaxis": {"domain": [0, 1], "anchor": "y3"},
                    "yaxis3": {"domain": [0, 0.3], "anchor": "x3", "nticks": 5},
                    "yaxis2": {"domain": [0.35, 0.65], "anchor": "x2", "nticks": 5},
                    "yaxis": {"domain": [0.7, 1], "nticks": 5},
                },
                "help": """\
A plot of the smoothly-varying crystal orientation parameters determined by
scan-varying refinement. By default, the crystal orientation matrix is
decomposed into rotations along the orthogonal axes of the laboratory frame.
Alternate axes may be chosen as parameters to the program. Small variations of
a few tenths of a degree are not uncommon. Larger changes may indicate a
problem with the model geometry. Monotonic drift along the axis closest to the
rotation axis may indicate that the goniometer rotation speed is recorded
incorrectly.
""",
            }
        }

        if len(dat) == 1:
            colors = color_repeats()
        else:
            colors = color_repeats(3)
        for ori in dat:
            d["scan_varying_orientation"]["data"].extend(
                [
                    {
                        "x": ori["phi"],
                        "y": ori["phi1"],
                        "type": "scatter",
                        "name": "Φ1 (°)",
                        "marker": {"color": next(colors)},
                    },
                    {
                        "x": ori["phi"],
                        "y": ori["phi2"],
                        "type": "scatter",
                        "name": "Φ2 (°)",
                        "xaxis": "x2",
                        "yaxis": "y2",
                        "marker": {"color": next(colors)},
                    },
                    {
                        "x": ori["phi"],
                        "y": ori["phi3"],
                        "type": "scatter",
                        "name": "Φ3 (°)",
                        "xaxis": "x3",
                        "yaxis": "y3",
                        "marker": {"color": next(colors)},
                    },
                ]
            )
        if not len(dat):
            return {}
        return d


class StrongSpotsAnalyser:
    """Analyse a list of strong spots."""

    def __init__(self, pixels_per_bin=10):
        self.pixels_per_bin = pixels_per_bin

        # Set the required fields
        self.required = ["xyzobs.px.value", "panel"]

    def __call__(self, rlist):
        """Analyse the strong spots."""
        # Check we have the required fields
        print("Analysing strong spots")
        if not ensure_required(rlist, self.required):
            return {"strong": {}}

        # Remove I_sigma <= 0
        if "intensity.sum.variance" in rlist:
            selection = rlist["intensity.sum.variance"] <= 0
            if selection.count(True) > 0:
                rlist.del_selected(selection)
                print(
                    " Removing %d reflections with variance <= 0"
                    % selection.count(True)
                )

        if "flags" in rlist:
            # Select only strong reflections
            Command.start(" Selecting only strong reflections")
            mask = rlist.get_flags(rlist.flags.strong)
            if mask.count(True) > 0:
                rlist = rlist.select(mask)
            Command.end(f" Selected {len(rlist)} strong reflections")

        x, y, z = rlist["xyzobs.px.value"].parts()
        self.nbinsx, self.nbinsy = tuple(
            int(math.ceil(i))
            for i in (
                flex.max(x) / self.pixels_per_bin,
                flex.max(y) / self.pixels_per_bin,
            )
        )

        # Look at distribution of spot counts
        d = self.spot_count_per_image(rlist)
        # Look at distribution of unindexed spots with detector position
        d.update(self.unindexed_count_xy(rlist))
        # Look at distribution of indexed spots with detector position
        d.update(self.indexed_count_xy(rlist))
        return {"strong": d}

    def spot_count_per_image(self, rlist):
        """Analyse the spot count per image."""
        x, y, z = rlist["xyzobs.px.value"].parts()
        max_z = int(math.ceil(flex.max(z)))

        indexed_sel = rlist.get_flags(rlist.flags.indexed)
        n_indexed = indexed_sel.count(True)

        if "imageset_id" in rlist:
            ids = rlist["imageset_id"]
        else:
            ids = rlist["id"]
        spot_count_per_image = []
        indexed_per_image = []
        for j in range(flex.max(ids) + 1):
            spot_count_per_image.append([])
            ids_sel = ids == j
            zsel = z.select(ids_sel)
            for i in range(max_z):
                sel = (zsel >= i) & (zsel < (i + 1))
                spot_count_per_image[j].append(sel.count(True))
            if n_indexed > 0:
                indexed_per_image.append([])
                zsel = z.select(ids_sel & indexed_sel)
                for i in range(max_z):
                    sel = (zsel >= i) & (zsel < (i + 1))
                    indexed_per_image[j].append(sel.count(True))

        d = {
            "spot_count_per_image": {
                "data": [],
                "layout": {
                    "title": "Spot count per image",
                    "xaxis": {"title": "Image"},
                    "yaxis": {"title": "Spot count", "rangemode": "tozero"},
                },
                "help": """\
A plot of the distribution of total and indexed spot count with respect to image
number. A drop off in spot count towards zero at the end of the scan may be
indicative of radiation damage, whereas a sudden fall followed by a sudden rise
in spot count may suggest that the crystal has moved out of the beam. Systematic
variations in spot count with image number may be a result of unit cell
dimensions, variations in volume of crystal intersecting the beam, or
diffraction anisotropy. Large separation between the total and indexed spot
count shows a significant number of unindexed reflections, which may be the
result of further, unidentified lattices, split reflections, reflections due to
ice rings, or poor spot-finding parameters.
""",
            }
        }
        for j in range(flex.max(ids) + 1):
            d["spot_count_per_image"]["data"].append(
                {
                    "x": list(range(len(spot_count_per_image[j]))),
                    "y": spot_count_per_image[j],
                    "type": "scatter",
                    "name": "#spots",
                    "opacity": 0.4,
                }
            )
            if n_indexed > 0:
                d["spot_count_per_image"]["data"].append(
                    {
                        "x": list(range(len(indexed_per_image[j]))),
                        "y": indexed_per_image[j],
                        "type": "scatter",
                        "name": "#indexed",
                        "opacity": 0.4,
                    }
                )

        if indexed_sel.count(True) > 0 and flex.max(rlist["id"]) > 0:
            # multiple lattices
            ids = rlist["id"]
            indexed_per_lattice_per_image = []
            for j in range(flex.max(ids) + 1):
                indexed_per_lattice_per_image.append([])
                zsel = z.select((ids == j) & indexed_sel)
                for i in range(max_z):
                    sel = (zsel >= i) & (zsel < (i + 1))
                    indexed_per_lattice_per_image[j].append(sel.count(True))

            d["indexed_per_lattice_per_image"] = {
                "data": [],
                "layout": {
                    "title": "Indexed reflections per lattice per image",
                    "xaxis": {"title": "Image"},
                    "yaxis": {"title": "Number of reflections", "rangemode": "tozero"},
                },
            }

            for j in range(flex.max(ids) + 1):
                d["indexed_per_lattice_per_image"]["data"].append(
                    {
                        "x": list(range(len(indexed_per_lattice_per_image[j]))),
                        "y": indexed_per_lattice_per_image[j],
                        "type": "scatter",
                        "name": "Lattice #%i" % (j + 1),
                        "opacity": 0.4,
                    }
                )

        return d

    def unindexed_count_xy(self, rlist):
        """Analyse the spot count in x/y."""
        x, y, z = rlist["xyzobs.px.value"].parts()

        indexed_sel = rlist.get_flags(rlist.flags.indexed)
        if indexed_sel.count(True) == 0 or indexed_sel.count(False) == 0:
            return {}

        x = x.select(~indexed_sel).as_numpy_array()
        y = y.select(~indexed_sel).as_numpy_array()

        H, xedges, yedges = np.histogram2d(x, y, bins=(self.nbinsx, self.nbinsy))

        return {
            "n_unindexed_vs_xy": {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": H.transpose().tolist(),
                        "type": "heatmap",
                        "name": "n_unindexed",
                        "colorbar": {
                            "title": "Number of reflections",
                            "titleside": "right",
                        },
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": "Number of unindexed reflections binned in X/Y",
                    "xaxis": {"domain": [0, 0.85], "title": "X", "showgrid": False},
                    "yaxis": {"title": "Y", "autorange": "reversed", "showgrid": False},
                    "width": 500,
                    "height": 450,
                },
            }
        }

    def indexed_count_xy(self, rlist):
        """Analyse the indexed spot count in x/y."""
        x, y, z = rlist["xyzobs.px.value"].parts()

        indexed_sel = rlist.get_flags(rlist.flags.indexed)
        if indexed_sel.count(True) == 0 or indexed_sel.count(False) == 0:
            return {}

        x = x.select(indexed_sel).as_numpy_array()
        y = y.select(indexed_sel).as_numpy_array()

        H, xedges, yedges = np.histogram2d(x, y, bins=(self.nbinsx, self.nbinsy))

        return {
            "n_indexed_vs_xy": {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": H.transpose().tolist(),
                        "type": "heatmap",
                        "name": "n_indexed",
                        "colorbar": {
                            "title": "Number of reflections",
                            "titleside": "right",
                        },
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": "Number of indexed reflections binned in X/Y",
                    "xaxis": {"domain": [0, 0.85], "title": "X", "showgrid": False},
                    "yaxis": {"title": "Y", "autorange": "reversed", "showgrid": False},
                    "width": 500,
                    "height": 450,
                },
            }
        }


class CentroidAnalyser:
    """Analyse the reflection centroids."""

    def __init__(self, grid_size=None, pixels_per_bin=10, centroid_diff_max=1.5):
        self.grid_size = grid_size
        self.pixels_per_bin = pixels_per_bin
        self.centroid_diff_max = centroid_diff_max

        # Set the required fields
        self.required = [
            "intensity.sum.value",
            "intensity.sum.variance",
            "xyzcal.px",
            "xyzobs.px.value",
            "xyzcal.mm",
            "xyzobs.mm.value",
        ]

    def __call__(self, rlist):
        """Analyse the reflection centroids."""
        # Check we have the required fields
        print("Analysing reflection centroids")
        if not ensure_required(rlist, self.required):
            return {"centroid": {}}

        # Remove I_sigma <= 0
        selection = rlist["intensity.sum.variance"] <= 0
        if selection.count(True) > 0:
            rlist.del_selected(selection)
            print(" Removing %d reflections with variance <= 0" % selection.count(True))

        # Remove partial reflections as their observed centroids won't be accurate
        if "partiality" in rlist:
            selection = rlist["partiality"] < 0.99
            if selection.count(True) > 0 and selection.count(True) < selection.size():
                rlist.del_selected(selection)
                print(" Removing %d partial reflections" % selection.count(True))

        # Select only integrated reflections
        Command.start(" Selecting only summation-integrated reflections")
        mask = rlist.get_flags(rlist.flags.integrated_sum)
        if mask.count(True) > 0:
            threshold = 10
            rlist = rlist.select(mask)
            Command.end(f" Selected {len(rlist)} summation-integrated reflections")
        else:
            # Select only those reflections used in refinement
            threshold = 0
            mask = rlist.get_flags(rlist.flags.used_in_refinement)
            rlist = rlist.select(mask)
            Command.end(f" Selected {len(rlist)} refined reflections")

        # Look at differences in calculated/observed position
        print(f" Analysing centroid differences with I/Sigma > {threshold}")
        d = self.centroid_diff_hist(rlist, threshold)
        print(f" Analysing centroid differences in x/y with I/Sigma > {threshold}")
        d.update(self.centroid_diff_xy(rlist, threshold))
        d.update(self.centroid_xy_xz_zy_residuals(rlist, threshold))
        print(f" Analysing centroid differences in z with I/Sigma > {threshold}")
        d.update(self.centroid_diff_z(rlist, threshold))
        print(f" Analysing centroid differences vs phi with I/Sigma > {threshold}")
        d.update(self.centroid_mean_diff_vs_phi(rlist, threshold))
        return {"centroid": d}

    def centroid_diff_hist(self, rlist, threshold):
        """Analyse the correlations."""
        I = rlist["intensity.sum.value"]
        I_sig = flex.sqrt(rlist["intensity.sum.variance"])
        I_over_S = I / I_sig
        mask = I_over_S > threshold
        if mask.count(True) == 0:
            return {}
        rlist = rlist.select(mask)
        assert len(rlist) > 0
        xc, yc, zc = rlist["xyzcal.px"].parts()
        xo, yo, zo = rlist["xyzobs.px.value"].parts()
        xd = xo - xc
        yd = yo - yc
        zd = zo - zc
        diff = flex.sqrt(xd * xd + yd * yd + zd * zd)
        hist = flex.histogram(diff, n_slots=20)

        d = {
            "centroid_difference_histogram": {
                "data": [
                    {
                        "x": list(hist.slot_centers()),
                        "y": list(hist.slots()),
                        "type": "bar",
                        "name": "centroid_differences",
                    }
                ],
                "layout": {
                    "title": "Difference between observed and calculated centroids",
                    "xaxis": {"title": "Difference in position (pixels)"},
                    "yaxis": {"title": "Number of reflections"},
                    "bargap": 0,
                },
            }
        }
        return d

    def centroid_diff_xy(self, rlist, threshold):
        """Look at the centroid difference in x, y"""
        I = rlist["intensity.sum.value"]
        I_sig = flex.sqrt(rlist["intensity.sum.variance"])
        I_over_S = I / I_sig
        mask = I_over_S > threshold
        if mask.count(True) == 0:
            return {}
        rlist = rlist.select(mask)
        assert len(rlist) > 0

        xc, yc, zc = rlist["xyzcal.px"].parts()
        xo, yo, zo = rlist["xyzobs.px.value"].parts()
        xd = xo - xc
        yd = yo - yc

        nbinsx, nbinsy = tuple(
            int(math.ceil(i))
            for i in (
                flex.max(xc) / self.pixels_per_bin,
                flex.max(yc) / self.pixels_per_bin,
            )
        )

        xc = xc.as_numpy_array()
        yc = yc.as_numpy_array()
        xd = xd.as_numpy_array()
        yd = yd.as_numpy_array()

        H, xedges, yedges = np.histogram2d(xc, yc, bins=(nbinsx, nbinsy))
        H1, xedges, yedges = np.histogram2d(xc, yc, bins=(nbinsx, nbinsy), weights=xd)
        H2, xedges, yedges = np.histogram2d(xc, yc, bins=(nbinsx, nbinsy), weights=yd)

        nonzeros = np.nonzero(H)
        z1 = np.empty(H.shape)
        z1[:] = np.NAN
        z1[nonzeros] = H1[nonzeros] / H[nonzeros]
        z2 = np.empty(H.shape)
        z2[:] = np.NAN
        z2[nonzeros] = H2[nonzeros] / H[nonzeros]

        return {
            "centroid_differences_x": {
                "data": [
                    {
                        "name": "centroid_differences_x",
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": z1.transpose().tolist(),
                        "type": "heatmap",
                        "colorbar": {
                            "title": "Difference in X position (pixels)",
                            "titleside": "right",
                        },
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": "Difference between observed and calculated centroids in X",
                    "xaxis": {"domain": [0, 0.85], "title": "X", "showgrid": False},
                    "yaxis": {"title": "Y", "autorange": "reversed", "showgrid": False},
                    "width": 500,
                    "height": 450,
                },
            },
            "centroid_differences_y": {
                "data": [
                    {
                        "name": "centroid_differences_y",
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": z2.transpose().tolist(),
                        "type": "heatmap",
                        "colorbar": {
                            "title": "Difference in Y position (pixels)",
                            "titleside": "right",
                        },
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": "Difference between observed and calculated centroids in Y",
                    "xaxis": {"domain": [0, 0.85], "title": "X", "showgrid": False},
                    "yaxis": {"title": "Y", "autorange": "reversed", "showgrid": False},
                    "width": 500,
                    "height": 450,
                },
            },
        }

    def centroid_diff_z(self, rlist, threshold):
        """Look at the centroid difference in z"""
        I = rlist["intensity.sum.value"]
        I_sig = flex.sqrt(rlist["intensity.sum.variance"])
        I_over_S = I / I_sig
        mask = I_over_S > threshold
        if mask.count(True) == 0:
            return {}
        rlist = rlist.select(mask)
        assert len(rlist) > 0
        xc, yc, zc = rlist["xyzcal.px"].parts()
        xo, yo, zo = rlist["xyzobs.px.value"].parts()
        zd = zo - zc

        if zd.all_approx_equal(zd[0]):
            # probably still images, no z residuals
            return {}

        H, xedges, yedges = np.histogram2d(zc, zd, bins=(100, 100))

        return {
            "centroid_differences_z": {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": H.transpose().tolist(),
                        "type": "heatmap",
                        "name": "centroid_differences_z",
                        "colorbar": {
                            "title": "Number of reflections",
                            "titleside": "right",
                        },
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": "Difference between observed and calculated centroids in Z",
                    "xaxis": {"title": "Z", "showgrid": False},
                    "yaxis": {
                        "title": "Difference in Z position (images)",
                        "showgrid": False,
                    },
                },
            }
        }

    def centroid_mean_diff_vs_phi(self, rlist, threshold):
        I = rlist["intensity.sum.value"]
        I_sig = flex.sqrt(rlist["intensity.sum.variance"])
        I_over_S = I / I_sig
        mask = I_over_S > threshold
        if mask.count(True) == 0:
            return {}
        rlist = rlist.select(mask)
        assert len(rlist) > 0

        xc, yc, zc = rlist["xyzcal.mm"].parts()
        xo, yo, zo = rlist["xyzobs.mm.value"].parts()

        dx = xc - xo
        dy = yc - yo
        dphi = (zc - zo) * RAD2DEG

        if dphi.all_approx_equal(dphi[0]):
            # probably still images, no z residuals
            return {}

        mean_residuals_x = flex.double()
        mean_residuals_y = flex.double()
        mean_residuals_phi = flex.double()
        rmsd_x = flex.double()
        rmsd_y = flex.double()
        rmsd_phi = flex.double()
        phi_obs_deg = RAD2DEG * zo
        phi = []

        for i_phi in range(
            int(math.floor(flex.min(phi_obs_deg))),
            int(math.ceil(flex.max(phi_obs_deg))),
        ):
            sel = (phi_obs_deg >= i_phi) & (phi_obs_deg < (i_phi + 1))
            if sel.count(True) == 0:
                continue
            mean_residuals_x.append(flex.mean(dx.select(sel)))
            mean_residuals_y.append(flex.mean(dy.select(sel)))
            mean_residuals_phi.append(flex.mean(dphi.select(sel)))
            rmsd_x.append(math.sqrt(flex.mean_sq(dx.select(sel))))
            rmsd_y.append(math.sqrt(flex.mean_sq(dy.select(sel))))
            rmsd_phi.append(math.sqrt(flex.mean_sq(dphi.select(sel))))
            phi.append(i_phi)

        d = {
            "centroid_mean_differences_vs_phi": {
                "data": [
                    {
                        "x": list(phi),
                        "y": list(mean_residuals_x),
                        "type": "scatter",
                        "name": "mean_dx",
                    },
                    {
                        "x": list(phi),
                        "y": list(mean_residuals_y),
                        "type": "scatter",
                        "name": "mean_dy",
                        "xaxis": "x2",
                        "yaxis": "y2",
                    },
                    {
                        "x": list(phi),
                        "y": list(mean_residuals_phi),
                        "type": "scatter",
                        "name": "mean_dphi",
                        "xaxis": "x3",
                        "yaxis": "y3",
                    },
                ],
                "layout": {
                    "title": "Difference between observed and calculated centroids vs phi",
                    "yaxis3": {"domain": [0, 0.266]},
                    # 'legend': {'traceorder': 'reversed'},
                    "xaxis3": {"anchor": "y3"},
                    "xaxis2": {"anchor": "y2"},
                    "yaxis2": {"domain": [0.366, 0.633]},
                    "yaxis": {"domain": [0.733, 1]},
                },
            },
            "centroid_rmsd_vs_phi": {
                "data": [
                    {
                        "x": list(phi),
                        "y": list(rmsd_x),
                        "type": "scatter",
                        "name": "rmsd_dx",
                    },
                    {
                        "x": list(phi),
                        "y": list(rmsd_y),
                        "type": "scatter",
                        "name": "rmsd_dy",
                        "xaxis": "x2",
                        "yaxis": "y2",
                    },
                    {
                        "x": list(phi),
                        "y": list(rmsd_phi),
                        "type": "scatter",
                        "name": "rmsd_dphi",
                        "xaxis": "x3",
                        "yaxis": "y3",
                    },
                ],
                "layout": {
                    "title": "RMSD between observed and calculated centroids vs phi",
                    "yaxis3": {"domain": [0, 0.266], "rangemode": "tozero"},
                    # 'legend': {'traceorder': 'reversed'},
                    "xaxis3": {"anchor": "y3"},
                    "xaxis2": {"anchor": "y2"},
                    "yaxis2": {"domain": [0.366, 0.633], "rangemode": "tozero"},
                    "yaxis": {"domain": [0.733, 1], "rangemode": "tozero"},
                },
            },
        }
        return d

    def centroid_xy_xz_zy_residuals(self, rlist, threshold):
        I = rlist["intensity.sum.value"]
        I_sig = flex.sqrt(rlist["intensity.sum.variance"])
        I_over_S = I / I_sig
        mask = I_over_S > threshold
        if mask.count(True) == 0:
            return {}
        rlist = rlist.select(mask)
        assert len(rlist) > 0

        xc, yc, zc = rlist["xyzcal.px"].parts()
        xo, yo, zo = rlist["xyzobs.px.value"].parts()
        dx = xc - xo
        dy = yc - yo
        dz = zc - zo

        is_stills = dz.all_approx_equal(dz[0])

        histx = flex.histogram(dx, n_slots=100)
        histy = flex.histogram(dy, n_slots=100)
        Hxy, xedges, yedges = np.histogram2d(dx, dy, bins=(50, 50))

        if not is_stills:
            histz = flex.histogram(dz, n_slots=100)
            Hzy, zedges, yedges = np.histogram2d(dz, dy, bins=(50, 50))
            Hxz, xedges, zedges = np.histogram2d(dx, dz, bins=(50, 50))

        density_hist_layout = {
            "showlegend": False,
            "autosize": False,
            "width": 600,
            "height": 550,
            "margin": {"t": 50},
            "hovermode": "closest",
            "bargap": 0,
            "xaxis": {
                "domain": [0, 0.85],
                "showgrid": False,
                "zeroline": True,
                "zerolinewidth": 2,
                "zerolinecolor": "#969696",
                "title": "X",
            },
            "yaxis": {
                "domain": [0, 0.85],
                "showgrid": False,
                "zeroline": True,
                "zerolinewidth": 2,
                "zerolinecolor": "#969696",
                "title": "Y",
            },
            "xaxis2": {
                "domain": [0.85, 1],
                "showgrid": False,
                "zeroline": False,
                "bargap": 0,
            },
            "yaxis2": {
                "domain": [0.85, 1],
                "showgrid": False,
                "zeroline": False,
                "bargap": 0,
            },
        }

        d = {
            "residuals_xy": {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": Hxy.transpose().tolist(),
                        "name": "density",
                        "colorscale": "Hot",
                        "reversescale": True,
                        "showscale": False,
                        "type": "contour",
                        "zsmooth": "best",
                    },
                    {
                        "x": list(histx.slot_centers()),
                        "y": list(histx.slots()),
                        "name": "dx histogram",
                        "marker": {"color": "rgb(102,0,0)"},
                        "yaxis": "y2",
                        "type": "bar",
                    },
                    {
                        "y": list(histy.slot_centers()),
                        "x": list(histy.slots()),
                        "name": "dy histogram",
                        "marker": {"color": "rgb(102,0,0)"},
                        "xaxis": "x2",
                        "type": "bar",
                        "orientation": "h",
                    },
                ],
                "layout": copy.deepcopy(density_hist_layout),
            }
        }
        d["residuals_xy"]["layout"]["title"] = "Centroid residuals in X and Y"

        if not is_stills:

            d["residuals_zy"] = {
                "data": [
                    {
                        "x": zedges.tolist(),
                        "y": yedges.tolist(),
                        "z": Hzy.transpose().tolist(),
                        "name": "density",
                        "colorscale": "Hot",
                        "reversescale": True,
                        "showscale": False,
                        "type": "contour",
                        "zsmooth": "best",
                    },
                    {
                        "x": list(histz.slot_centers()),
                        "y": list(histz.slots()),
                        "name": "dz histogram",
                        "marker": {"color": "rgb(102,0,0)"},
                        "yaxis": "y2",
                        "type": "bar",
                    },
                    {
                        "y": list(histy.slot_centers()),
                        "x": list(histy.slots()),
                        "name": "dy histogram",
                        "marker": {"color": "rgb(102,0,0)"},
                        "xaxis": "x2",
                        "type": "bar",
                        "orientation": "h",
                    },
                ],
                "layout": copy.deepcopy(density_hist_layout),
            }
            d["residuals_zy"]["layout"]["title"] = "Centroid residuals in Z and Y"
            d["residuals_zy"]["layout"]["xaxis"]["title"] = "Z"

            d["residuals_xz"] = {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": zedges.tolist(),
                        "z": Hxz.transpose().tolist(),
                        "name": "density",
                        "colorscale": "Hot",
                        "reversescale": True,
                        "showscale": False,
                        "type": "contour",
                        "zsmooth": "best",
                    },
                    {
                        "x": list(histx.slot_centers()),
                        "y": list(histx.slots()),
                        "name": "dx histogram",
                        "marker": {"color": "rgb(102,0,0)"},
                        "yaxis": "y2",
                        "type": "bar",
                    },
                    {
                        "y": list(histz.slot_centers()),
                        "x": list(histz.slots()),
                        "name": "dz histogram",
                        "marker": {"color": "rgb(102,0,0)"},
                        "xaxis": "x2",
                        "type": "bar",
                        "orientation": "h",
                    },
                ],
                "layout": copy.deepcopy(density_hist_layout),
            }
            d["residuals_xz"]["layout"]["title"] = "Centroid residuals in X and Z"
            d["residuals_xz"]["layout"]["yaxis"]["title"] = "Z"

        return d


class IntensityAnalyser:
    """Analyse the intensities."""

    def __init__(self, grid_size=None, pixels_per_bin=10):

        self.grid_size = grid_size
        self.pixels_per_bin = pixels_per_bin

        # Set the required fields
        self.required = ["intensity.sum.value", "intensity.sum.variance", "xyzcal.px"]

    def __call__(self, rlist):
        """Analyse the reflection centroids."""
        # FIXME Do the same and a comparison for intensity.prf

        # Check we have the required fields
        print("Analysing reflection intensities")
        if not ensure_required(rlist, self.required):
            return {"intensity": {}}

        selection = rlist["intensity.sum.variance"] <= 0
        if selection.count(True) > 0:
            rlist.del_selected(selection)
            print(" Removing %d reflections with variance <= 0" % selection.count(True))

        selection = rlist["intensity.sum.value"] <= 0
        if selection.count(True) > 0:
            rlist.del_selected(selection)
            print(
                " Removing %d reflections with intensity <= 0" % selection.count(True)
            )

        # Select only integrated reflections
        Command.start(" Selecting only integrated reflections")
        mask = rlist.get_flags(rlist.flags.integrated, all=False)
        if mask.count(True) == 0:
            return {"intensity": {}}

        rlist = rlist.select(mask)
        Command.end(f" Selected {len(rlist)} integrated reflections")

        x, y, z = rlist["xyzcal.px"].parts()
        self.nbinsx, self.nbinsy = tuple(
            int(math.ceil(i))
            for i in (
                flex.max(x) / self.pixels_per_bin,
                flex.max(y) / self.pixels_per_bin,
            )
        )

        print(" Analysing distribution of I/Sigma")
        d = self.i_over_s_hist(rlist)
        print(" Analysing distribution of I/Sigma vs xy")
        d.update(self.i_over_s_vs_xy(rlist, "sum"))
        if "intensity.prf.value" in rlist:
            print(" Analysing distribution of I/Sigma vs xy")
            d.update(self.i_over_s_vs_xy(rlist, "prf"))
        print(" Analysing distribution of I/Sigma vs z")
        d.update(self.i_over_s_vs_z(rlist))
        print(" Analysing distribution of partialities")
        d.update(self.partiality_hist(rlist))
        print(" Analysing QE map")
        d.update(self.qe_xy(rlist))
        return {"intensity": d}

    def i_over_s_hist(self, rlist):
        """Analyse the correlations."""
        I = rlist["intensity.sum.value"]
        I_sig = flex.sqrt(rlist["intensity.sum.variance"])
        I_over_S = I / I_sig
        log_I_over_S = flex.log10(I_over_S)
        hist = flex.histogram(log_I_over_S, n_slots=20)

        return {
            "log_i_over_sigma_histogram": {
                "data": [
                    {
                        "x": list(hist.slot_centers()),
                        "y": list(hist.slots()),
                        "type": "bar",
                        "name": "log_i_over_sigma",
                    }
                ],
                "layout": {
                    "title": "Log I/σ(I) histogram",
                    "xaxis": {"title": "Log I/σ(I)"},
                    "yaxis": {"title": "Number of reflections"},
                    "bargap": 0,
                },
            }
        }

    def i_over_s_vs_xy(self, rlist, intensity_type):
        """Plot I/Sigma vs X/Y"""

        I_sig = flex.sqrt(rlist[f"intensity.{intensity_type}.variance"])
        I = rlist[f"intensity.{intensity_type}.value"]
        sel = (I_sig > 0) & (I > 0)
        rlist = rlist.select(sel)
        I = I.select(sel)
        I_sig = I_sig.select(sel)
        I_over_S = I / I_sig
        x, y, z = rlist["xyzcal.px"].parts()

        H, xedges, yedges = np.histogram2d(
            x.as_numpy_array(), y.as_numpy_array(), bins=(self.nbinsx, self.nbinsy)
        )
        H1, xedges, yedges = np.histogram2d(
            x.as_numpy_array(),
            y.as_numpy_array(),
            bins=(self.nbinsx, self.nbinsy),
            weights=flex.log10(I_over_S).as_numpy_array(),
        )

        nonzeros = np.nonzero(H)
        z = np.empty(H.shape)
        z[:] = np.NAN
        z[nonzeros] = H1[nonzeros] / H[nonzeros]

        return {
            f"i_over_sigma_{intensity_type}_vs_xy": {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": z.transpose().tolist(),
                        "zmin": -1,
                        "zauto": False,
                        "type": "heatmap",
                        "name": f"i_over_sigma_{intensity_type}",
                        "colorbar": {"title": "Log I/σ(I)", "titleside": "right"},
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": f"Distribution of I({intensity_type})/σ vs X/Y",
                    "xaxis": {"domain": [0, 0.85], "title": "X", "showgrid": False},
                    "yaxis": {"title": "Y", "autorange": "reversed", "showgrid": False},
                    "width": 500,
                    "height": 450,
                },
            }
        }

    def i_over_s_vs_z(self, rlist):
        """Plot I/Sigma vs Z."""

        I = rlist["intensity.sum.value"]
        I_sig = flex.sqrt(rlist["intensity.sum.variance"])
        I_over_S = I / I_sig
        x, y, z = rlist["xyzcal.px"].parts()

        H, xedges, yedges = np.histogram2d(
            z.as_numpy_array(), flex.log10(I_over_S).as_numpy_array(), bins=(100, 100)
        )

        return {
            "i_over_sigma_vs_z": {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": H.transpose().tolist(),
                        "type": "heatmap",
                        "name": "i_over_sigma",
                        "colorbar": {
                            "title": "Number of reflections",
                            "titleside": "right",
                        },
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": "Distribution of I/σ(I) vs Z",
                    "xaxis": {"title": "Z", "showgrid": False},
                    "yaxis": {"title": "Log I/σ(I)", "showgrid": False},
                },
            }
        }

    def partiality_hist(self, rlist):
        """Analyse the partialities."""
        partiality = rlist["partiality"]
        hist = flex.histogram(partiality.select(partiality > 0), 0, 1, n_slots=20)

        return {
            "partiality_histogram": {
                "data": [
                    {
                        "x": list(hist.slot_centers()),
                        "y": list(hist.slots()),
                        "type": "bar",
                        "name": "partiality",
                    }
                ],
                "layout": {
                    "title": "Partiality histogram",
                    "xaxis": {"title": "Partiality", "range": [0, 1]},
                    "yaxis": {"title": "Number of reflections"},
                    "bargap": 0,
                },
            }
        }

    def qe_xy(self, rlist):
        """Look at the QE map in x, y"""

        if "qe" not in rlist:
            return {}

        assert len(rlist) > 0
        qe = rlist["qe"]

        xc, yc, zc = rlist["xyzcal.px"].parts()

        nbinsx, nbinsy = tuple(
            int(math.ceil(i))
            for i in (
                flex.max(xc) / self.pixels_per_bin,
                flex.max(yc) / self.pixels_per_bin,
            )
        )

        xc = xc.as_numpy_array()
        yc = yc.as_numpy_array()
        qe = qe.as_numpy_array()

        H, xedges, yedges = np.histogram2d(xc, yc, bins=(nbinsx, nbinsy))
        H1, xedges, yedges = np.histogram2d(xc, yc, bins=(nbinsx, nbinsy), weights=qe)

        nonzeros = np.nonzero(H)
        z1 = np.empty(H.shape)
        z1[:] = np.NAN
        z1[nonzeros] = H1[nonzeros] / H[nonzeros]

        return {
            "qe_map": {
                "data": [
                    {
                        "name": "qe_map",
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": z1.transpose().tolist(),
                        "type": "heatmap",
                        "colorbar": {"title": "QE", "titleside": "right"},
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": "Calculated Quantum Efficiency",
                    "xaxis": {"domain": [0, 0.85], "title": "X", "showgrid": False},
                    "yaxis": {"title": "Y", "autorange": "reversed", "showgrid": False},
                    "width": 500,
                    "height": 450,
                },
            }
        }


class ZScoreAnalyser:
    """
    Analyse the distribution of intensity z-scores.

    z-scores are calculated as z = (I - <I>) / σ, where I is intensity,
    <I> is a weighted mean and σ is the measurement error in intensity,
    as calculated by dials.util.intensity_explorer.
    """

    def __init__(self):
        # Set the required fields
        self.required = [
            "miller_index",
            "intensity.sum.value",
            "intensity.sum.variance",
            "xyzobs.px.value",
        ]

    def __call__(self, rlist, experiments):
        """
        :param rlist: A reflection table containing, at the least, the following
        fields:
          * ``miller_index``
          * ``intensity.sum.value``
          * ``intensity.sum.variance``
          * ``xyzobs.px.value``
        :type rlist: `dials_array_family_flex_ext.reflection_table`
        :param experiments: An experiment list with space group information.
        :type experiments: `dxtbx.model.ExperimentList`
        :return: A dictionary describing several Plotly plots.
        :rtype:`dict`
        """

        from dials.util.intensity_explorer import IntensityDist

        # Check we have the required fields
        print("Analysing reflection intensities")
        if not ensure_required(rlist, self.required):
            return {}

        selection = rlist["intensity.sum.variance"] <= 0
        if selection.count(True) > 0:
            rlist.del_selected(selection)
            print(" Removing %d reflections with variance <= 0" % selection.count(True))

        selection = rlist["intensity.sum.value"] <= 0
        if selection.count(True) > 0:
            rlist.del_selected(selection)
            print(
                " Removing %d reflections with intensity <= 0" % selection.count(True)
            )

        self.z_score_data = IntensityDist(rlist, experiments).rtable

        print(" Analysing distribution of intensity z-scores")
        d = self.z_score_hist()
        print(" Constructing normal probability plot of intensity z-scores")
        d.update(self.normal_probability_plot())
        print(" Plotting intensity z-scores against multiplicity")
        d.update(self.z_vs_multiplicity())
        print(" Plotting time series of intensity z-scores")
        d.update(self.z_time_series())
        print(" Plotting intensity z-scores versus weighted mean intensity")
        d.update(self.z_vs_I())
        print(" Plotting intensity z-scores versus I/sigma")
        d.update(self.z_vs_I_over_sigma())

        return d

    def z_score_hist(self):
        """
        Plot a histogram of z-scores.

        :return: A dictionary describing a Plotly plot.
        :rtype:`dict`
        """

        z_scores = self.z_score_data["intensity.z_score"]
        hist = flex.histogram(z_scores, -10, 10, 100)

        return {
            "z_score_histogram": {
                "data": [
                    {
                        "x": hist.slot_centers().as_numpy_array().tolist(),
                        "y": hist.slots().as_numpy_array().tolist(),
                        "type": "bar",
                        "name": "z_hist",
                    }
                ],
                "layout": {
                    "title": "Histogram of z-scores",
                    "xaxis": {"title": "z-score", "range": [-10, 10]},
                    "yaxis": {"title": "Number of reflections"},
                    "bargap": 0,
                },
            }
        }

    def normal_probability_plot(self):
        """
        Display a normal probability plot of the z-scores of the intensities.

        :return: A dictionary describing a Plotly plot.
        :rtype:`dict`
        """

        z_scores = self.z_score_data["intensity.z_score"]
        osm = self.z_score_data["intensity.order_statistic_medians"]

        return {
            "normal_probability_plot": {
                "data": [
                    {
                        "x": osm.as_numpy_array().tolist(),
                        "y": z_scores.as_numpy_array().tolist(),
                        "type": "scattergl",
                        "mode": "markers",
                        "name": "Data",
                    },
                    {
                        "x": [-5, 5],
                        "y": [-5, 5],
                        "type": "scatter",
                        "mode": "lines",
                        "name": "z = m",
                    },
                ],
                "layout": {
                    "title": "Normal probability plot",
                    "xaxis": {"title": "z-score order statistic medians, m"},
                    "yaxis": {
                        "title": "Ordered z-score responses, z",
                        "range": [-10, 10],
                    },
                },
            }
        }

    def z_vs_multiplicity(self):
        """
        Plot z-score as a function of multiplicity of observation.

        :return: A dictionary describing a Plotly plot.
        :rtype:`dict`
        """

        multiplicity = self.z_score_data["multiplicity"]
        z_scores = self.z_score_data["intensity.z_score"]

        return {
            "z_score_vs_multiplicity": {
                "data": [
                    {
                        "x": multiplicity.as_numpy_array().tolist(),
                        "y": z_scores.as_numpy_array().tolist(),
                        "type": "scattergl",
                        "mode": "markers",
                        "name": "Data",
                    }
                ],
                "layout": {
                    "title": "z-scores versus multiplicity",
                    "xaxis": {"title": "Multiplicity"},
                    "yaxis": {"title": "z-score"},
                },
            }
        }

    def z_time_series(self):
        """
        Plot a crude time-series of z-scores, with image number serving as a
        proxy for time.

        :return: A dictionary describing a Plotly plot.
        :rtype:`dict`
        """

        batch_number = self.z_score_data["xyzobs.px.value"].parts()[2]
        z_scores = self.z_score_data["intensity.z_score"]

        return {
            "z_score_time_series": {
                "data": [
                    {
                        "x": batch_number.as_numpy_array().tolist(),
                        "y": z_scores.as_numpy_array().tolist(),
                        "type": "scattergl",
                        "mode": "markers",
                        "name": "Data",
                    }
                ],
                "layout": {
                    "title": "z-scores versus image number",
                    "xaxis": {"title": "Image number"},
                    "yaxis": {"title": "z-score"},
                },
            }
        }

    def z_vs_I(self):
        """
        Plot z-scores versus intensity.

        :return: A dictionary describing a Plotly plot.
        :rtype:`dict`
        """

        intensity = self.z_score_data["intensity.mean.value"]
        z_scores = self.z_score_data["intensity.z_score"]

        return {
            "z_score_vs_I": {
                "data": [
                    {
                        "x": intensity.as_numpy_array().tolist(),
                        "y": z_scores.as_numpy_array().tolist(),
                        "type": "scattergl",
                        "mode": "markers",
                        "name": "Data",
                    }
                ],
                "layout": {
                    "title": "z-scores versus weighted mean intensity",
                    "xaxis": {"title": "Intensity (photon count)", "type": "log"},
                    "yaxis": {"title": "z-score"},
                },
            }
        }

    def z_vs_I_over_sigma(self):
        """
        Plot z-scores versus intensity.

        :return: A dictionary describing a Plotly plot.
        :rtype:`dict`
        """

        i_over_sigma = (
            self.z_score_data["intensity.mean.value"]
            / self.z_score_data["intensity.mean.std_error"]
        )
        z_scores = self.z_score_data["intensity.z_score"]

        return {
            "z_score_vs_I_over_sigma": {
                "data": [
                    {
                        "x": i_over_sigma.as_numpy_array().tolist(),
                        "y": z_scores.as_numpy_array().tolist(),
                        "type": "scattergl",
                        "mode": "markers",
                        "name": "Data",
                    }
                ],
                "layout": {
                    "title": "z-scores versus I/σ",
                    "xaxis": {"title": "I/σ", "type": "log"},
                    "yaxis": {"title": "z-score"},
                },
            }
        }


class ReferenceProfileAnalyser:
    """Analyse the reference profiles."""

    def __init__(self, grid_size=None, pixels_per_bin=10):
        self.grid_size = grid_size
        self.pixels_per_bin = pixels_per_bin

        # Set the required fields
        self.required = [
            "intensity.prf.value",
            "intensity.prf.variance",
            "xyzcal.px",
            "profile.correlation",
        ]

    def __call__(self, rlist):
        """Analyse the reference profiles."""
        # Check we have the required fields
        print("Analysing reference profiles")
        if not ensure_required(rlist, self.required):
            return {"reference": {}}

        # Select only integrated reflections
        Command.start(" Selecting only integrated reflections")
        mask = rlist.get_flags(rlist.flags.integrated)
        if mask.count(True) == 0:
            return {"reference": {}}

        rlist = rlist.select(mask)
        Command.end(f" Selected {len(rlist)} integrated reflections")

        x, y, z = rlist["xyzcal.px"].parts()
        self.nbinsx, self.nbinsy = tuple(
            int(math.ceil(i))
            for i in (
                flex.max(x) / self.pixels_per_bin,
                flex.max(y) / self.pixels_per_bin,
            )
        )

        # Analyse distribution of reference spots
        print(" Analysing reference profile distribution vs x/y")
        d = self.reference_xy(rlist)
        print(" Analysing reference profile distribution vs z")
        d.update(self.reference_z(rlist))

        # Look at correlations between profiles
        def correlations(filename, rlist):
            """Call for reference spots and all reflections."""

            print(" Analysing reflection profile correlations")
            d = self.reflection_corr_hist(rlist, filename)

            print(" Analysing reflection profile correlations vs x/y")
            d.update(self.reflection_corr_vs_xy(rlist, filename))

            print(" Analysing reflection profile correlations vs z")
            d.update(self.reflection_corr_vs_z(rlist, filename))

            print(" Analysing reflection profile correlations vs I/Sigma")
            d.update(self.reflection_corr_vs_ios(rlist, filename))

            return d

        mask = rlist.get_flags(rlist.flags.reference_spot)
        d.update(self.reflection_correlations_vs_resolution(rlist))
        d.update(correlations("reference", rlist.select(mask)))
        d.update(correlations("reflection", rlist))

        return {"reference": d}

    def reflection_correlations_vs_resolution(self, rlist):
        """Analyse the distribution of reference profiles."""

        print(" Analysing reflection correlations vs resolution")
        from dials.algorithms.spot_finding.per_image_analysis import binner_d_star_cubed

        profile_correlation = rlist["profile.correlation"]
        d_spacings = rlist["d"]
        binner = binner_d_star_cubed(d_spacings)
        bin_centres = flex.double()
        ccs = flex.double()
        for d_bin in binner.bins:
            d_min = d_bin.d_min
            d_max = d_bin.d_max
            ds3_min = 1 / d_min**3
            ds3_max = 1 / d_max**3
            ds3_centre = (ds3_max - ds3_min) / 2 + ds3_min
            d_centre = 1 / ds3_centre ** (1 / 3)
            sel = (d_spacings < d_max) & (d_spacings >= d_min)
            if sel.count(True) == 0:
                continue
            bin_centres.append(d_centre)
            ccs.append(flex.mean(profile_correlation.select(sel)))

        d_star_sq_bins = uctbx.d_as_d_star_sq(bin_centres)

        def d_star_sq_to_d_ticks(d_star_sq, nticks):
            min_d_star_sq = min(d_star_sq)
            dstep = (max(d_star_sq) - min_d_star_sq) / nticks
            tickvals = [min_d_star_sq + (i * dstep) for i in range(nticks)]
            ticktext = [f"{uctbx.d_star_sq_as_d(dsq):.2f}" for dsq in tickvals]
            return tickvals, ticktext

        tickvals, ticktext = d_star_sq_to_d_ticks(d_star_sq_bins, nticks=5)

        return {
            "reflection_cc_vs_resolution": {
                "data": [
                    {
                        "x": list(d_star_sq_bins),  # d_star_sq
                        "y": list(ccs),
                        "type": "scatter",
                        "name": "profile_correlation",
                    }
                ],
                "layout": {
                    "title": "Reflection correlations vs resolution",
                    "xaxis": {
                        "title": "Resolution (Å)",
                        "tickvals": tickvals,
                        "ticktext": ticktext,
                    },
                    "yaxis": {
                        "title": "Correlation with reference profile",
                        "range": [0, 1],
                    },
                },
            }
        }

    def reference_xy(self, rlist):
        """Analyse the distribution of reference profiles."""
        mask = rlist.get_flags(rlist.flags.reference_spot)
        rlist = rlist.select(mask)
        x, y, z = rlist["xyzcal.px"].parts()

        H, xedges, yedges = np.histogram2d(
            x.as_numpy_array(), y.as_numpy_array(), bins=(self.nbinsx, self.nbinsy)
        )

        return {
            "n_reference_profiles_vs_xy": {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": H.transpose().tolist(),
                        "type": "heatmap",
                        "name": "n_reference_profiles",
                        "colorbar": {
                            "title": "Number of reflections",
                            "titleside": "right",
                        },
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": "Reference profiles binned in X/Y",
                    "xaxis": {"domain": [0, 0.85], "title": "X", "showgrid": False},
                    "yaxis": {"title": "Y", "autorange": "reversed", "showgrid": False},
                    "width": 500,
                    "height": 450,
                },
            }
        }

    def reference_z(self, rlist):
        """Analyse the distribution of reference profiles."""
        x, y, z = rlist["xyzcal.px"].parts()
        hist = flex.histogram(z, n_slots=20)

        return {
            "n_reference_profiles_vs_z": {
                "data": [
                    {
                        "x": list(hist.slot_centers()),
                        "y": list(hist.slots()),
                        "type": "bar",
                        "name": "n_reference_profiles",
                    }
                ],
                "layout": {
                    "title": "Reference profiles binned in Z",
                    "xaxis": {"title": "Z"},
                    "yaxis": {"title": "Number of reflections"},
                    "bargap": 0,
                },
            }
        }

    def reflection_corr_hist(self, rlist, filename):
        """Analyse the correlations."""
        corr = rlist["profile.correlation"]
        hist = flex.histogram(corr, n_slots=20)

        return {
            f"{filename}_correlations_histogram": {
                "data": [
                    {
                        "x": list(hist.slot_centers()),
                        "y": list(hist.slots()),
                        "type": "bar",
                        "name": f"{filename}_correlations",
                    }
                ],
                "layout": {
                    "title": f"{filename.capitalize()} correlations histogram",
                    "xaxis": {"title": "Correlation with reference profile"},
                    "yaxis": {"title": "Number of reflections"},
                    "bargap": 0,
                },
            }
        }

    def reflection_corr_vs_xy(self, rlist, filename):
        """Analyse the correlations."""

        corr = rlist["profile.correlation"]
        x, y, z = rlist["xyzcal.px"].parts()

        H, xedges, yedges = np.histogram2d(
            x.as_numpy_array(), y.as_numpy_array(), bins=(self.nbinsx, self.nbinsy)
        )
        H1, xedges, yedges = np.histogram2d(
            x.as_numpy_array(),
            y.as_numpy_array(),
            bins=(self.nbinsx, self.nbinsy),
            weights=corr.as_numpy_array(),
        )

        nonzeros = np.nonzero(H)
        z = np.empty(H.shape)
        z[:] = np.NAN
        z[nonzeros] = H1[nonzeros] / H[nonzeros]

        return {
            f"{filename}_correlations_xy": {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": z.transpose().tolist(),
                        "type": "heatmap",
                        "name": f"{filename}_correlations",
                        "colorbar": {
                            "title": "Correlation with reference profile",
                            "titleside": "right",
                        },
                        "colorscale": "Viridis",
                        "zmin": 0,
                        "zmax": 1,
                    }
                ],
                "layout": {
                    "title": f"{filename.capitalize()} correlations binned in X/Y",
                    "xaxis": {"domain": [0, 0.85], "title": "X", "showgrid": False},
                    "yaxis": {"title": "Y", "autorange": "reversed", "showgrid": False},
                    "width": 500,
                    "height": 450,
                },
            }
        }

    def reflection_corr_vs_z(self, rlist, filename):
        """Analyse the correlations."""

        corr = rlist["profile.correlation"]
        x, y, z = rlist["xyzcal.px"].parts()

        H, xedges, yedges = np.histogram2d(
            z.as_numpy_array(), corr.as_numpy_array(), bins=(100, 100)
        )

        return {
            f"{filename}_correlations_vs_z": {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": H.transpose().tolist(),
                        "type": "heatmap",
                        "name": f"{filename}_correlations",
                        "colorbar": {
                            "title": "Number of reflections",
                            "titleside": "right",
                        },
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": f"{filename.capitalize()} correlations vs Z",
                    "xaxis": {"title": "Z", "showgrid": False},
                    "yaxis": {
                        "title": "Correlation with reference profile",
                        "showgrid": False,
                    },
                },
            }
        }

    def reflection_corr_vs_ios(self, rlist, filename):
        """Analyse the correlations."""

        corr = rlist["profile.correlation"]
        I = rlist["intensity.prf.value"]
        I_sig = flex.sqrt(rlist["intensity.prf.variance"])
        mask = I_sig > 0
        I = I.select(mask)
        I_sig = I_sig.select(mask)
        corr = corr.select(mask)
        I_over_S = I / I_sig
        mask = I_over_S > 0.1
        I_over_S = I_over_S.select(mask)
        corr = corr.select(mask)

        H, xedges, yedges = np.histogram2d(
            flex.log10(I_over_S).as_numpy_array(),
            corr.as_numpy_array(),
            bins=(100, 100),
        )

        return {
            f"{filename}_correlations_vs_ios": {
                "data": [
                    {
                        "x": xedges.tolist(),
                        "y": yedges.tolist(),
                        "z": H.transpose().tolist(),
                        "type": "heatmap",
                        "name": f"{filename}_correlations",
                        "colorbar": {
                            "title": "Number of reflections",
                            "titleside": "right",
                        },
                        "colorscale": "Viridis",
                    }
                ],
                "layout": {
                    "title": f"{filename.capitalize()} correlations vs Log I/σ(I)",
                    "xaxis": {"title": "Log I/σ(I)", "showgrid": False},
                    "yaxis": {
                        "title": "Correlation with reference profile",
                        "showgrid": False,
                    },
                },
            }
        }


from dials.algorithms.merging.merge import MergingStatisticsData


def merging_stats_data(reflections, experiments):
    reflections["intensity"] = reflections["intensity.scale.value"]
    reflections["variance"] = reflections["intensity.scale.variance"]

    reflections = (
        reflections.split_by_experiment_id()
    )  # needed to get correct batch plots
    scaled_array = scaled_data_as_miller_array(reflections, experiments)
    stats_data = MergingStatisticsData(reflections, experiments, scaled_array)
    stats, anom_stats = merging_stats_from_scaled_array(
        scaled_array,
        n_bins=20,
        use_internal_variance=False,
    )
    stats_data.merging_statistics_result = stats
    stats_data.anom_merging_statistics_result = anom_stats
    stats_plots = make_merging_stats_plots(stats_data, run_xtriage_analysis=True)
    intensity_plots = stats_plots["anom_plots"]
    intensity_plots.update(
        i_over_sig_i_vs_i_plot(scaled_array.data(), scaled_array.sigmas())
    )
    return (
        stats_plots["resolution_plots"],
        stats_plots["xtriage_output"],
        stats_plots["misc_plots"],
        intensity_plots,
        stats_plots["batch_plots"],
        stats_plots["image_range_tables"],
        stats_plots["scaling_tables"],
    )


class Analyser:
    """Helper class to do all the analysis."""

    def __init__(self, params, grid_size=None, centroid_diff_max=1.5):
        """Setup the analysers."""
        self.params = params
        self.analysers = [
            StrongSpotsAnalyser(pixels_per_bin=self.params.pixels_per_bin),
            CentroidAnalyser(
                grid_size=grid_size,
                pixels_per_bin=self.params.pixels_per_bin,
                centroid_diff_max=centroid_diff_max,
            ),
            IntensityAnalyser(
                grid_size=grid_size, pixels_per_bin=self.params.pixels_per_bin
            ),
            ReferenceProfileAnalyser(
                grid_size=grid_size, pixels_per_bin=self.params.pixels_per_bin
            ),
        ]

    def __call__(self, rlist=None, experiments=None):
        """Do all the analysis."""
        json_data = {}

        if rlist is not None:
            for analyse in self.analysers:
                result = analyse(copy.deepcopy(rlist))
                if result is not None:
                    json_data.update(result)
        else:
            json_data.update(
                {"strong": {}, "centroid": {}, "intensity": {}, "reference": {}}
            )

        crystal_table = None
        expt_geom_table = None
        if experiments is not None:
            analyse = ScanVaryingCrystalAnalyser(self.params.orientation_decomposition)
            json_data.update(analyse(experiments))
            crystal_table, expt_geom_table = self.experiments_table(experiments)
            json_data.update(
                {
                    "scaling_model": {},
                    "xtriage_output": {},
                    "resolution_graphs": {},
                    "batch_graphs": {},
                    "misc_graphs": {},
                    "scaled_intensity_graphs": {},
                    "scaling_tables": [],
                    "image_range_tables": [],
                }
            )
            if any(experiments.scaling_models()):
                print("Analysing scaling model")
                d = {}
                for i, model in enumerate(experiments.scaling_models()):
                    if model is not None:
                        scaling_model_plots = plot_scaling_models(model)
                        if scaling_model_plots:
                            for name, plot in scaling_model_plots.items():
                                d.update({name + "_" + str(i): plot})
                json_data["scaling_model"] = d

            if rlist:
                if "inverse_scale_factor" in rlist:
                    print("Calculating and generating merging statistics plots")
                    try:
                        (
                            resolution_plots,
                            xtriage_output,
                            misc_plots,
                            scaled_intensity_plots,
                            batch_plots,
                            image_range_tables,
                            scaling_tables,
                        ) = merging_stats_data(rlist, experiments)
                    except DialsMergingStatisticsError as e:
                        print(e)
                    else:
                        json_data["resolution_graphs"] = resolution_plots
                        json_data["xtriage_output"] = xtriage_output
                        json_data["batch_graphs"] = batch_plots
                        json_data["misc_graphs"] = misc_plots
                        json_data["scaled_intensity_graphs"] = scaled_intensity_plots
                        json_data["scaling_tables"] = scaling_tables
                        json_data["image_range_tables"] = image_range_tables

        if self.params.output.html is not None:

            from jinja2 import ChoiceLoader, Environment, PackageLoader

            loader = ChoiceLoader(
                [
                    PackageLoader("dials", "templates"),
                    PackageLoader("dials", "static", encoding="utf-8"),
                ]
            )
            env = Environment(loader=loader)

            graphs = json_data

            import libtbx.load_env

            static_dir = libtbx.env.find_in_repositories("dials/static")

            if self.params.output.external_dependencies == "local":
                template = env.get_template("report_local_dep.html")
            elif self.params.output.external_dependencies == "embed":
                template = env.get_template("report_embed_dep.html")
            else:
                template = env.get_template("report.html")
            html = template.render(
                page_title="DIALS analysis report",
                scan_varying_graphs=graphs["scan_varying"],
                scaling_model_graphs=graphs["scaling_model"],
                xtriage_output=graphs["xtriage_output"],
                resolution_plots=graphs["resolution_graphs"],
                batch_graphs=graphs["batch_graphs"],
                image_range_tables=graphs["image_range_tables"],
                misc_plots=graphs["misc_graphs"],
                scaled_intensity_plots=graphs["scaled_intensity_graphs"],
                strong_graphs=graphs["strong"],
                centroid_graphs=graphs["centroid"],
                intensity_graphs=graphs["intensity"],
                reference_graphs=graphs["reference"],
                crystal_table=crystal_table,
                geometry_table=expt_geom_table,
                scaling_tables=graphs["scaling_tables"],
                static_dir=static_dir,
            )

            print(f"Writing html report to: {self.params.output.html}")
            with open(self.params.output.html, "wb") as f:
                f.write(html.encode("utf-8", "xmlcharrefreplace"))

        if self.params.output.json is not None:
            print(f"Writing json data to: {self.params.output.json}")
            with open(self.params.output.json, "w", encoding="utf-8") as f:
                json.dump(json_data, f)

    def experiments_table(self, experiments):
        assert experiments is not None

        crystal_table = []
        expt_geom_table = []

        latex_vector_template = (
            r"$$\left( \begin{array}{rrr} %5.4f & %5.4f & %5.4f \end{array} \right)$$"
        )
        latex_matrix_template = "\n".join(
            (
                r"$$\left( \begin{array}{rrr}",
                r"%5.4f & %5.4f & %5.4f \\",
                r"%5.4f & %5.4f & %5.4f \\",
                r"%5.4f & %5.4f & %5.4f \end{array} \right)$$",
            )
        )
        latex_unit_cell_template = r"$$\left( \begin{array}{cccccc} %5.3f & %5.3f & %5.3f & %5.3f & %5.3f & %5.3f \end{array}\right)$$"

        for expt in experiments:
            for panel_id, panel in enumerate(expt.detector):
                expt_geom_table.append(
                    (
                        "<strong>Panel %i</strong>:" % (panel_id + 1),
                        "Pixel size (mm):",
                        "%.4f, %.4f" % panel.get_pixel_size(),
                        "Image size (pixels):",
                        "%i, %i" % panel.get_image_size(),
                    )
                )
                expt_geom_table.append(
                    (
                        "",
                        "Trusted range:",
                        "%g, %g" % panel.get_trusted_range(),
                        "Thickness (mm):",
                        f"{panel.get_thickness():g}",
                    )
                )
                expt_geom_table.append(
                    (
                        "",
                        "Material:",
                        f"{panel.get_material()}",
                        "μ:",
                        f"{panel.get_mu():g}",
                    )
                )
                expt_geom_table.append(
                    (
                        "",
                        "Fast axis:",
                        latex_vector_template % panel.get_fast_axis(),
                        "Slow axis:",
                        latex_vector_template % panel.get_slow_axis(),
                    )
                )
                expt_geom_table.append(
                    (
                        "",
                        "Origin:",
                        latex_vector_template % panel.get_origin(),
                        "Distance (mm)",
                        f"{panel.get_distance():.4f}",
                    )
                )
                if len(expt.detector) == 1:
                    try:
                        # does the beam intersect with the panel?
                        panel.get_beam_centre(expt.beam.get_s0())
                    except RuntimeError:
                        continue
                    else:
                        expt_geom_table.append(
                            (
                                "",
                                "Max resolution (corners) (Å):",
                                "%.2f"
                                % panel.get_max_resolution_at_corners(
                                    expt.beam.get_s0()
                                ),
                                "Max resolution (inscribed circle) (Å):",
                                "%.2f"
                                % panel.get_max_resolution_ellipse(expt.beam.get_s0()),
                            )
                        )

            if expt.scan is not None:
                expt_geom_table.append(
                    (
                        "<strong>Scan:</strong>",
                        "Image range:",
                        "%i, %i" % expt.scan.get_image_range(),
                        "Oscillation:",
                        "%.2f&deg;, %+.2f&deg;/frame" % expt.scan.get_oscillation(),
                    )
                )

            if expt.goniometer is not None:
                expt_geom_table.append(
                    (
                        "<strong>Goniometer:</strong>",
                        "Rotation axis:",
                        latex_vector_template % expt.goniometer.get_rotation_axis(),
                    )
                )
                expt_geom_table.append(
                    (
                        "",
                        "Fixed rotation:",
                        latex_matrix_template % expt.goniometer.get_fixed_rotation(),
                        "Setting rotation:",
                        latex_matrix_template % expt.goniometer.get_setting_rotation(),
                    )
                )

            if expt.crystal is not None:
                uc = expt.crystal.get_unit_cell().parameters()
                sgi = expt.crystal.get_space_group().info()
                umat = latex_matrix_template % expt.crystal.get_U()
                bmat = latex_matrix_template % expt.crystal.get_B()
                amat = latex_matrix_template % expt.crystal.get_A()
                crystal_table.append(
                    (
                        "<strong>Crystal:</strong>",
                        "Space group:",
                        sgi.symbol_and_number(),
                        "Unit cell:",
                        latex_unit_cell_template % uc,
                    )
                )
                crystal_table.append(
                    ("", "U matrix:", f"{umat}", "B matrix:", f"{bmat}")
                )
                crystal_table.append(("", "A = UB:", f"{amat}"))
                if expt.crystal.num_scan_points > 0:
                    abc = flex.vec3_double()
                    angles = flex.vec3_double()
                    for n in range(expt.crystal.num_scan_points):
                        (
                            a,
                            b,
                            c,
                            alpha,
                            beta,
                            gamma,
                        ) = expt.crystal.get_unit_cell_at_scan_point(n).parameters()
                        abc.append((a, b, c))
                        angles.append((alpha, beta, gamma))
                    a, b, c = abc.mean()
                    alpha, beta, gamma = angles.mean()
                    mean_uc = uctbx.unit_cell((a, b, c, alpha, beta, gamma))
                    crystal_table.append(
                        (
                            "",
                            "A sampled at %i scan points"
                            % expt.crystal.num_scan_points,
                            "",
                            "Average unit cell:",
                            latex_unit_cell_template % mean_uc.parameters(),
                        )
                    )

        # ensure all the rows are the same length
        for table in (expt_geom_table, crystal_table):
            for i_row in range(len(table)):
                while len(table[i_row]) < 5:
                    table[i_row] = list(table[i_row]) + [""]

        return crystal_table, expt_geom_table


class Script:
    """A class to encapsulate the script."""

    def __init__(self):
        """Initialise the script."""
        from dials.util.options import ArgumentParser

        # Create the parser
        usage = "usage: dials.report [options] observations.refl"
        self.parser = ArgumentParser(
            usage=usage,
            phil=phil_scope,
            read_reflections=True,
            read_experiments=True,
            check_format=False,
            epilog=help_message,
        )
        dials.util.log.print_banner()

    def run(self, args=None):
        """Run the script."""
        from dials.util.options import reflections_and_experiments_from_files

        # Parse the command line arguments
        params, options = self.parser.parse_args(args, show_diff_phil=True)

        # Show the help
        if len(params.input.reflections) != 1 and not len(params.input.experiments):
            self.parser.print_help()
            exit(0)

        reflections, experiments = reflections_and_experiments_from_files(
            params.input.reflections, params.input.experiments
        )

        # Analyse the reflections
        analyse = Analyser(
            params,
            grid_size=params.grid_size,
            centroid_diff_max=params.centroid_diff_max,
        )
        if len(reflections):
            reflections = reflections[0]
        else:
            reflections = None

        analyse(reflections, experiments)


@show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

import math


def test():
    from scitbx.array_family import flex

    from dials.algorithms.image.fill_holes import simple_fill

    mask = flex.bool(flex.grid(100, 100), True)
    data = flex.double(flex.grid(100, 100), True)

    for j in range(100):
        for i in range(100):
            data[j, i] = 10 + j * 0.01 + i * 0.01
            if math.sqrt((j - 50) ** 2 + (i - 50) ** 2) <= 10.5:
                mask[j, i] = False
                data[j, i] = 0

    result = simple_fill(data, mask)
    known = data.as_1d().select(mask.as_1d())
    filled = result.as_1d().select(~mask.as_1d())
    assert flex.max(filled) <= flex.max(known)
    assert flex.min(filled) >= flex.min(known)



from __future__ import annotations

from math import pi

from dxtbx.model import Crystal
from rstbx.symmetry.constraints.parameter_reduction import symmetrize_reduce_enlarge
from scitbx import matrix

from dials.algorithms.refinement.parameterisation.crystal_parameters import (
    CrystalOrientationParameterisation,
    CrystalUnitCellParameterisation,
)
from dials.algorithms.refinement.refinement_helpers import (
    get_fd_gradients,
    random_param_shift,
)


def test():
    import random
    import textwrap

    from cctbx.uctbx import unit_cell
    from libtbx.test_utils import approx_equal

    def random_direction_close_to(vector):
        return vector.rotate_around_origin(
            matrix.col((random.random(), random.random(), random.random())).normalize(),
            random.gauss(0, 1.0),
            deg=True,
        )

    # make a random P1 crystal and parameterise it
    a = random.uniform(10, 50) * random_direction_close_to(matrix.col((1, 0, 0)))
    b = random.uniform(10, 50) * random_direction_close_to(matrix.col((0, 1, 0)))
    c = random.uniform(10, 50) * random_direction_close_to(matrix.col((0, 0, 1)))
    xl = Crystal(a, b, c, space_group_symbol="P 1")

    xl_op = CrystalOrientationParameterisation(xl)
    xl_ucp = CrystalUnitCellParameterisation(xl)

    null_mat = matrix.sqr((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))

    # compare analytical and finite difference derivatives
    an_ds_dp = xl_op.get_ds_dp()
    fd_ds_dp = get_fd_gradients(xl_op, [1.0e-6 * pi / 180] * 3)
    for e, f in zip(an_ds_dp, fd_ds_dp):
        assert approx_equal((e - f), null_mat, eps=1.0e-6)

    an_ds_dp = xl_ucp.get_ds_dp()
    fd_ds_dp = get_fd_gradients(xl_ucp, [1.0e-7] * xl_ucp.num_free())
    for e, f in zip(an_ds_dp, fd_ds_dp):
        assert approx_equal((e - f), null_mat, eps=1.0e-6)

    # random initial orientations with a random parameter shift at each
    attempts = 100
    for i in range(attempts):

        # make a random P1 crystal and parameterise it
        a = random.uniform(10, 50) * random_direction_close_to(matrix.col((1, 0, 0)))
        b = random.uniform(10, 50) * random_direction_close_to(matrix.col((0, 1, 0)))
        c = random.uniform(10, 50) * random_direction_close_to(matrix.col((0, 0, 1)))
        xl = Crystal(a, b, c, space_group_symbol="P 1")
        xl_op = CrystalOrientationParameterisation(xl)
        xl_uc = CrystalUnitCellParameterisation(xl)

        # apply a random parameter shift to the orientation
        p_vals = xl_op.get_param_vals()
        p_vals = random_param_shift(
            p_vals, [1000 * pi / 9, 1000 * pi / 9, 1000 * pi / 9]
        )
        xl_op.set_param_vals(p_vals)

        # compare analytical and finite difference derivatives
        xl_op_an_ds_dp = xl_op.get_ds_dp()
        xl_op_fd_ds_dp = get_fd_gradients(xl_op, [1.0e-5 * pi / 180] * 3)

        # apply a random parameter shift to the unit cell. We have to
        # do this in a way that is respectful to metrical constraints,
        # so don't modify the parameters directly; modify the cell
        # constants and extract the new parameters
        cell_params = xl.get_unit_cell().parameters()
        cell_params = random_param_shift(cell_params, [1.0] * 6)
        new_uc = unit_cell(cell_params)
        newB = matrix.sqr(new_uc.fractionalization_matrix()).transpose()
        S = symmetrize_reduce_enlarge(xl.get_space_group())
        S.set_orientation(orientation=newB)
        X = S.forward_independent_parameters()
        xl_uc.set_param_vals(X)

        xl_uc_an_ds_dp = xl_ucp.get_ds_dp()

        # now doing finite differences about each parameter in turn
        xl_uc_fd_ds_dp = get_fd_gradients(xl_ucp, [1.0e-7] * xl_ucp.num_free())

        for j in range(3):
            assert approx_equal(
                (xl_op_fd_ds_dp[j] - xl_op_an_ds_dp[j]), null_mat, eps=1.0e-6
            ), textwrap.dedent(
                """\
        Failure in try {i}
        failure for parameter number {j}
        of the orientation parameterisation
        with fd_ds_dp =
        {fd}
        and an_ds_dp =
        {an}
        so that difference fd_ds_dp - an_ds_dp =
        {diff}
        """
            ).format(
                i=i,
                j=j,
                fd=xl_op_fd_ds_dp[j],
                an=xl_op_an_ds_dp[j],
                diff=xl_op_fd_ds_dp[j] - xl_op_an_ds_dp[j],
            )

        for j in range(xl_ucp.num_free()):
            assert approx_equal(
                (xl_uc_fd_ds_dp[j] - xl_uc_an_ds_dp[j]), null_mat, eps=1.0e-6
            ), textwrap.dedent(
                """\
        Failure in try {i}
        failure for parameter number {j}
        of the unit cell parameterisation
        with fd_ds_dp =
        {fd}
        and an_ds_dp =
        {an}
        so that difference fd_ds_dp - an_ds_dp =
        {diff}
        """
            ).format(
                i=i,
                j=j,
                fd=xl_uc_fd_ds_dp[j],
                an=xl_uc_an_ds_dp[j],
                diff=xl_uc_fd_ds_dp[j] - xl_uc_an_ds_dp[j],
            )



from __future__ import annotations

from dials.extensions.glm_background_ext import GLMBackgroundExt
from dials.extensions.simple_background_ext import SimpleBackgroundExt


class AutoBackgroundExt:
    """An extension class to automatically choose a background algorithm.
    If experiments all have counting detectors then the glm algorithm is
    chosen. Otherwise, the simple algorithm is chosen."""

    name = "Auto"

    default = True

    def __init__(self, params, experiments):
        """
        Initialise the algorithm.

        :param params: The input parameters
        :param experiments: The list of experiments
        """

        # Guess at whether we have a counting detector, based on gains and
        # detector types (https://github.com/dials/dials/issues/706)
        detectors = experiments.detectors()
        gains = [p.get_gain() for d in detectors for p in d]
        pnl_types = [p.get_type() for d in detectors for p in d]
        counting_detector = True
        if [g == 1.0 for g in gains].count(False) != 0:
            counting_detector = False
        if ["PAD" in t for t in pnl_types].count(False) != 0:
            counting_detector = False

        if counting_detector:
            ext = GLMBackgroundExt(params, experiments)
        else:
            ext = SimpleBackgroundExt(params, experiments)

        self._algorithm = ext._algorithm

    def compute_background(self, reflections, image_volume=None):
        """
        Compute the background.

        :param reflections: The list of reflections
        """
        return self._algorithm.compute_background(
            reflections, image_volume=image_volume
        )



from __future__ import annotations

import datetime

import procrunner
import pytest

# Conflicts with downstream dependencies causing warnings. Temporarily
# ignore this test failing. Remove after 2022-11-01
if datetime.date.today() < datetime.date(2022, 11, 1):
    timed_xfail = pytest.mark.xfail(
        reason="Downstream dependencies causing warnings on some platforms. Give some time for resolution."
    )
else:
    timed_xfail = lambda x: x


@timed_xfail
def test(dials_data, tmp_path):

    mtz_file = (
        dials_data("lysozyme_electron_diffraction", pathlib=True) / "refmac_final.mtz"
    )
    result = procrunner.run(
        ["dials.plot_Fo_vs_Fc", f"hklin={mtz_file}"], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("Fo_vs_Fc.pdf").is_file()
    assert "|Fe| = 42.0" in result.stdout.decode()



from __future__ import annotations

import os
import pathlib

import pytest

from cctbx import sgtbx, uctbx
from dxtbx.serialize import load

import dials.command_line.cosym as dials_cosym
from dials.algorithms.symmetry.cosym._generate_test_data import (
    generate_experiments_reflections,
)
from dials.array_family import flex
from dials.util import Sorry


@pytest.mark.parametrize(
    "space_group,engine", [(None, "scitbx"), ("P 1", "scipy"), ("P 4", "scipy")]
)
def test_cosym(dials_data, run_in_tmp_path, space_group, engine):
    mcp = dials_data("multi_crystal_proteinase_k", pathlib=True)
    args = ["space_group=" + str(space_group), "seed=0", f"engine={engine}"]
    for i in [1, 2, 3, 4, 5, 7, 8, 10]:
        args.append(os.fspath(mcp / f"experiments_{i}.json"))
        args.append(os.fspath(mcp / f"reflections_{i}.pickle"))
    dials_cosym.run(args=args)
    assert pathlib.Path("symmetrized.refl").is_file()
    assert pathlib.Path("symmetrized.expt").is_file()
    experiments = load.experiment_list("symmetrized.expt", check_format=False)
    if space_group is None:
        assert (
            experiments[0].crystal.get_space_group().type().lookup_symbol() == "P 4 2 2"
        )
    else:
        assert (
            experiments[0].crystal.get_space_group().type().lookup_symbol()
            == space_group
        )
    joint_reflections = flex.reflection_table.from_file("symmetrized.refl")
    # check that there are 8 unique id and imageset_ids, and that these
    # correctly correspond to each experiment
    assert len(set(joint_reflections["id"])) == 8
    assert len(set(joint_reflections["imageset_id"])) == 8
    for id_ in range(8):
        sel = joint_reflections["id"] == id_
        assert set(joint_reflections["imageset_id"].select(sel)) == {id_}


def test_cosym_partial_dataset(dials_data, run_in_tmp_path):
    """Test how cosym handles partial/bad datasets."""
    mcp = dials_data("multi_crystal_proteinase_k", pathlib=True)
    args = []
    for i in [1, 2]:
        args.append(os.fspath(mcp / f"experiments_{i}.json"))
        args.append(os.fspath(mcp / f"reflections_{i}.pickle"))
    # Make one dataset that will be removed in prefiltering
    r = flex.reflection_table.from_file(mcp / "reflections_8.pickle")
    r["partiality"] = flex.double(r.size(), 0.1)
    r.as_file("renamed.refl")
    args.append("renamed.refl")
    args.append(os.fspath(mcp / "experiments_8.json"))
    # Add another good dataset at the end of the input list
    args.append(os.fspath(mcp / "experiments_10.json"))
    args.append(os.fspath(mcp / "reflections_10.pickle"))
    args.append("threshold=None")

    dials_cosym.run(args=args)
    assert pathlib.Path("symmetrized.refl").is_file()
    assert pathlib.Path("symmetrized.expt").is_file()
    experiments = load.experiment_list("symmetrized.expt", check_format=False)
    assert len(experiments) == 3


def test_cosym_resolution_filter_excluding_datasets(dials_data, run_in_tmp_path):
    mcp = dials_data("multi_crystal_proteinase_k", pathlib=True)
    args = ["space_group=P4", "seed=0", "d_min=20.0"]
    for i in [1, 2, 3, 4, 5, 7, 8, 10]:
        args.append(os.fspath(mcp / f"experiments_{i}.json"))
        args.append(os.fspath(mcp / f"reflections_{i}.pickle"))
    dials_cosym.run(args=args)
    assert pathlib.Path("symmetrized.refl").is_file()
    assert pathlib.Path("symmetrized.expt").is_file()

    joint_reflections = flex.reflection_table.from_file("symmetrized.refl")
    # check that there are 8 unique id and imageset_ids, and that these
    # correctly correspond to each experiment
    assert set(joint_reflections["id"]) == {0, 1, 2, 3, 4, 5, 6}
    assert len(set(joint_reflections["imageset_id"])) == 7
    for id_ in range(7):
        sel = joint_reflections["id"] == id_
        assert set(joint_reflections["imageset_id"].select(sel)) == {id_}


def test_cosym_partial_dataset_raises_sorry(dials_data, run_in_tmp_path, capsys):
    """Test how cosym handles partial/bad datasets."""
    mcp = dials_data("multi_crystal_proteinase_k", pathlib=True)
    args = ["renamed.refl", os.fspath(mcp / "experiments_8.json")]
    r2 = flex.reflection_table.from_file(mcp / "reflections_10.pickle")
    r2["partiality"] = flex.double(r2.size(), 0.1)
    r2.as_file("renamed2.refl")
    args.append("renamed2.refl")
    args.append(os.fspath(mcp / "experiments_10.json"))

    with pytest.raises(Sorry):
        dials_cosym.run(args=args)


def test_cosym_with_reference(dials_data, run_in_tmp_path):
    """Test indexing ambiguity resolution against a reference. Use
    SSX example data as have reference data for these."""
    ssx = dials_data("cunir_serial_processed", pathlib=True)
    refls = ssx / "integrated.refl"
    expts = ssx / "integrated.expt"
    reference = dials_data("cunir_serial", pathlib=True) / "2bw4-sf.cif"

    args = ["d_min=2.0", str(refls), str(expts), f"reference={reference}"]
    dials_cosym.run(args=args)
    assert pathlib.Path("symmetrized.refl").is_file()
    assert pathlib.Path("symmetrized.expt").is_file()
    experiments = load.experiment_list("symmetrized.expt", check_format=False)
    assert len(experiments) == 5  # check we have the right number of expts


@pytest.mark.parametrize(
    (
        "space_group",
        "unit_cell",
        "dimensions",
        "sample_size",
        "use_known_space_group",
        "use_known_lattice_group",
    ),
    [
        ("P2", None, None, 10, False, False),
        ("P3", None, None, 20, False, False),
        ("I23", None, 2, 10, False, False),
        ("P422", (79, 79, 37, 90, 90, 90), None, 10, True, False),
        ("P321", (59.39, 59.39, 28.35, 90, 90, 120), None, 5, False, False),
    ],
)
def test_synthetic(
    space_group,
    unit_cell,
    dimensions,
    sample_size,
    use_known_space_group,
    use_known_lattice_group,
    run_in_tmp_path,
):
    space_group = sgtbx.space_group_info(space_group).group()
    if unit_cell is not None:
        unit_cell = uctbx.unit_cell(unit_cell)
    experiments, reflections, _ = generate_experiments_reflections(
        space_group=space_group,
        unit_cell=unit_cell,
        unit_cell_volume=10000,
        sample_size=sample_size,
        map_to_p1=True,
        d_min=1.5,
    )

    experiments.as_json("tmp.expt")
    expt_file = "tmp.expt"
    joint_table = flex.reflection_table()
    for r in reflections:
        joint_table.extend(r)
    joint_table.as_file("tmp.refl")
    refl_file = "tmp.refl"

    args = [
        expt_file,
        refl_file,
        "output.experiments=symmetrized.expt",
        "output.reflections=symmetrized.refl",
        "output.html=cosym.html",
        "output.json=cosym.json",
    ]

    if use_known_space_group:
        args.append(f"space_group={space_group.info()}")
    if use_known_lattice_group:
        args.append(f"lattice_group={space_group.info()}")
    if dimensions is not None:
        args.append(f"dimensions={dimensions}")

    dials_cosym.run(args=args)
    assert pathlib.Path("symmetrized.refl").is_file()
    assert pathlib.Path("symmetrized.expt").is_file()
    assert pathlib.Path("cosym.html").is_file()
    assert pathlib.Path("cosym.json").is_file()
    cosym_expts = load.experiment_list("symmetrized.expt", check_format=False)
    assert len(cosym_expts) == len(experiments)
    for expt in cosym_expts:
        if unit_cell is not None:
            assert expt.crystal.get_unit_cell().parameters() == pytest.approx(
                unit_cell.parameters()
            )
        if (
            str(expt.crystal.get_space_group().info()) == "P 6 2 2"
            and str(space_group.info()) == "P 3 2 1"
        ):
            # This is fine
            continue
        assert str(expt.crystal.get_space_group().info()) == str(space_group.info())
        assert expt.crystal.get_space_group() == space_group



from __future__ import annotations

from dials.util import tabulate


class Array:
    """
    A class to represent an array
    """

    def __init__(self):
        """
        Initialise the array
        """
        self.name = ""
        self.title = ""
        self.data = None

    def as_dict(self):
        """
        Return as a dictionary

        :return: The dictionary
        """

        return {
            "title": self.title,
            "shape": self.data.all(),
            "data": list(self.data),
        }

    def as_str(self, prefix=""):
        """
        Return as a string

        :return: The string
        """
        return ""


class Table:
    """
    A class to represent a table
    """

    def __init__(self):
        """
        Initialize the table
        """
        self.name = ""
        self.title = ""
        self.cols = []
        self.rows = []

    def as_dict(self):
        """
        Return as a dictionary

        :return: The dictionary
        """
        cols = {col[0]: col[1] for col in self.cols}
        rows = [{self.cols[j][0]: r for j, r in enumerate(row)} for row in self.rows]

        # Create the output
        return {
            "title": self.title,
            "cols": cols,
            "rows": rows,
        }

    def as_str(self, prefix=""):
        """
        Return the table as a string

        :return: The string
        """
        rows = [[col[1] for col in self.cols]]
        for i, row in enumerate(self.rows):
            rows.append([str(x) for x in row])
        text = [prefix + self.title, tabulate(rows, headers="firstrow"), ""]
        return "\n".join(text)


class Report:
    """
    A class to represent the report
    """

    def __init__(self):
        """
        Initialize the tables
        """
        self.tables = []
        self.arrays = []

    def add_array(self, array):
        """
        Add an array to the report

        :param array: The array to add
        """
        self.arrays.append(array)

    def add_table(self, table):
        """
        Add a table to the report

        :param table: The table to add
        """
        self.tables.append(table)

    def combine(self, other):
        """
        Combine two reports

        :param other: The other report
        """
        self.tables.extend(other.tables)
        self.arrays.extend(other.arrays)

    def as_dict(self):
        """
        Return the report as a dictionary

        :return: The dictionary
        """
        return {
            "tables": {table.name: table.as_dict() for table in self.tables},
            "arrays": {array.name: array.as_dict() for array in self.arrays},
        }

    def as_str(self, prefix=""):
        """
        Return the report as a string

        :return: The string
        """
        return "\n".join([table.as_str(prefix) for table in self.tables])

    def as_json(self):
        """
        Save the report as a json file

        :return: The json string
        """
        import json

        return json.dumps(self.as_dict(), indent=2)

    def as_xml(self):
        """
        Save the report as an xml file

        :return: The XML string
        """
        from xml.dom import minidom

        # Get the XML implementation
        impl = minidom.getDOMImplementation()

        # Create the XML document
        doc = impl.createDocument(None, None, None)

        # Create the document root
        root = doc.createElement("DIALS")

        # Function to process objects
        def process(root, obj):
            if isinstance(obj, dict):
                for key, value in obj.items():
                    root.appendChild(process(doc.createElement(key), value))
            elif isinstance(obj, list) or isinstance(obj, tuple):
                for i, value in enumerate(obj):
                    root.appendChild(process(doc.createElement("%d" % i), value))
            else:
                root.appendChild(doc.createTextNode(str(obj)))
            return root

        # Process the dictionary and append
        doc.appendChild(process(root, self.as_dict()))

        # Return the XML document
        return doc.toprettyxml(indent="  ")

    def as_file(self, filename):
        """
        Export as a file (either json or xml depending on extension

        :param filename: The filename
        """
        from os.path import splitext

        ext = splitext(filename)[1]
        with open(filename, "w") as outfile:
            if ext == ".json":
                outfile.write(self.as_json())
            elif ext == ".xml":
                outfile.write(self.as_xml())
            else:
                raise RuntimeError("Filename must be *.xml or *.json")



"""
Tools to run cluster processing using DRMAA
"""

from __future__ import annotations

import multiprocessing
import os
import pickle
import sys


class InputWriter:
    """
    A class to write the input files
    """

    def __init__(self, directory, function, iterable):
        """
        Save the function and iterable
        """
        self.directory = directory
        self.function = function
        self.iterable = iterable

    def __call__(self):
        """
        Call this to write input files
        """

        for i, item in enumerate(self.iterable, start=1):
            with open(os.path.join(self.directory, "%d.input" % i), "wb") as outfile:
                pickle.dump(
                    (self.function, item), outfile, protocol=pickle.HIGHEST_PROTOCOL
                )


def cluster_map(func, iterable, callback=None, nslots=1, njobs=1, job_category="low"):
    """
    A function to map stuff on cluster using drmaa

    :param func: The function to call
    :param iterable: The iterable to pass to each function call
    :param callback: A callback function when each job completes
    :param nslots: The number of processes to request per cluster node
    """
    import tempfile

    import drmaa

    # Set the working directory and make sure it exists
    # This will be where all the input/output files associated with the cluster
    # submission will go. For each job there will be a file:
    #  - ${INDEX}.input     The pickled input to the job
    #  - ${INDEX}.output    The pickled output from the job
    #  - ${INDEX}.stdout    The stdout from the job
    #  - ${INDEX}.stderr    The stderr from the job
    cwd = tempfile.mkdtemp(prefix="dials_cluster_map_", dir=os.getcwd())

    # Start outputting the input files in a separate process
    process = multiprocessing.Process(target=InputWriter(cwd, func, iterable))
    process.start()
    process.join()

    # Start the drmaa session
    with drmaa.Session() as s:

        # Create the job template
        jt = s.createJobTemplate()
        jt.remoteCommand = "cluster.dials.exec"
        jt.args = [cwd]
        jt.jobName = "dials"
        jt.joinFiles = True
        jt.jobEnvironment = os.environ
        jt.workingDirectory = cwd
        jt.outputPath = ":" + os.path.join(
            cwd, f"{drmaa.JobTemplate.PARAMETRIC_INDEX}.stdout"
        )
        jt.errorPath = ":" + os.path.join(
            cwd, f"{drmaa.JobTemplate.PARAMETRIC_INDEX}.stderr"
        )
        jt.jobCategory = job_category

        # FIXME Currently no portable way of specifying this
        # In order to select a cluster node with N cores
        # we have to use the native specification. This will work
        # on SGE but may not work for other queuing systems
        # This will set the NSLOTS environment variable
        jt.nativeSpecification = "-pe smp %d -tc %d" % (nslots, njobs)

        N = len(list(iterable))
        try:

            # Submit the array job
            joblist = s.runBulkJobs(jt, 1, N, 1)

            # For each item, load the result and process
            result = []
            for i, jobid in enumerate(joblist, start=1):
                s.wait(jobid, drmaa.Session.TIMEOUT_WAIT_FOREVER)
                with open(os.path.join(cwd, "%d.stdout" % i)) as infile:
                    sys.stdout.write("".join(infile.readlines()))
                with open(os.path.join(cwd, "%d.output" % i), "rb") as infile:
                    r = pickle.load(infile)
                if isinstance(r, Exception):
                    raise r
                if callback is not None:
                    callback(r)
                result.append(r)

            # Make sure the process has finished
            # process.join()

            # Delete job template
            s.deleteJobTemplate(jt)

        except KeyboardInterrupt:

            # Delete the jobs
            s.control(
                drmaa.Session.JOB_IDS_SESSION_ALL, drmaa.JobControlAction.TERMINATE
            )

            # Delete job template
            s.deleteJobTemplate(jt)

            # Re-raise exception
            raise

    # Return the result
    return result


if __name__ == "__main__":
    from dials.util.cluster_func_test import func

    print(
        cluster_map(
            func,
            list(range(100)),
            nslots=4,
            # callback=print,
        )
    )



from __future__ import annotations

from dials_algorithms_polygon_spatial_interpolation_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "Match",
    "grid_to_irregular_grid",
    "irregular_grid_to_grid",
    "regrid_grid_to_irregular_grid",
    "regrid_irregular_grid_to_grid",
)



from __future__ import annotations

import copy

import libtbx.phil
from scitbx import matrix

from dials.array_family import flex

phil_scope = libtbx.phil.parse(
    """
reverse_phi = False
  .type = bool
  .optional = True
crystal_frame = False
  .type = bool
  .optional = True
beam_centre_panel = None
  .type = int
  .help = "Panel number for the beam centre"
beam_centre = None
  .type = floats(size=2)
  .help = "Fast, slow beam centre coordinates (mm)."
d_min = None
  .type = float(value_min=0.0)
z_min = None
  .type = float
z_max = None
  .type = float
n_min = 0
  .type = int
  .help = "Minimum size of spot in pixels"
n_max = 1000000
  .type = int
  .help = "Maximum size of spot in pixels"
partiality_min = None
  .type = float
partiality_max = None
  .type = float
display = *all unindexed indexed integrated
  .type = choice
outlier_display = outliers inliers
  .type = choice
experiment_ids = None
  .type = ints(value_min=-1)
black_background = True
  .type = bool
  .help = "Switch between black or white background"
"""
)


class Render3d:
    def __init__(self, settings=None):
        self.reflections = None
        if settings is None:
            self.settings = phil_scope.fetch().extract()
        else:
            self.settings = settings
        self.goniometer_orig = None
        self.viewer = None

    def load_models(self, experiments, reflections):
        self.experiments = experiments
        self.reflections_input = reflections
        if self.experiments[0].goniometer is not None:
            self.viewer.set_rotation_axis(
                self.experiments[0].goniometer.get_rotation_axis()
            )
        self.viewer.set_beam_vector(self.experiments[0].beam.get_s0())
        if self.settings.beam_centre is None:
            try:
                (
                    self.settings.beam_centre_panel,
                    self.settings.beam_centre,
                ) = self.experiments[0].detector.get_ray_intersection(
                    self.experiments[0].beam.get_s0()
                )
            except RuntimeError:
                pass
        else:
            self.set_beam_centre(
                self.settings.beam_centre_panel, self.settings.beam_centre
            )
        crystals = [
            expt.crystal for expt in self.experiments if expt.crystal is not None
        ]
        if crystals:
            # the points are scaled by 100 so must do that here too
            vecs = [
                [
                    matrix.col(l)
                    for l in (100 * matrix.sqr(c.get_A()))
                    .transpose()
                    .as_list_of_lists()
                ]
                for c in crystals
            ]
            self.viewer.set_reciprocal_lattice_vectors(vecs)
            vecs = [
                [
                    matrix.col(l)
                    for l in (100 * matrix.sqr(c.get_B()))
                    .transpose()
                    .as_list_of_lists()
                ]
                for c in crystals
            ]
            self.viewer.set_reciprocal_crystal_vectors(vecs)
        self.map_points_to_reciprocal_space()
        self.set_points()

    def map_points_to_reciprocal_space(self):
        # 155 handle data from predictions *only* if that is what we have
        calculated = "xyzobs.px.value" not in self.reflections_input
        self.reflections = copy.deepcopy(self.reflections_input)
        experiments = copy.deepcopy(self.experiments)

        if not calculated:
            self.reflections.centroid_px_to_mm(experiments)

        if self.settings.reverse_phi:
            for expt in experiments:
                expt.goniometer.set_rotation_axis(
                    [-i for i in expt.goniometer.get_rotation_axis()]
                )

        self.reflections.map_centroids_to_reciprocal_space(
            experiments,
            calculated=calculated,
            crystal_frame=self.settings.crystal_frame,
        )

    def set_points(self):
        reflections = self.reflections

        if "miller_index" in reflections:
            if "flags" not in reflections:
                reflections.set_flags(
                    reflections["miller_index"] != (0, 0, 0), reflections.flags.indexed
                )
                reflections["id"].set_selected(
                    reflections["miller_index"] == (0, 0, 0), -1
                )
                reflections.set_flags(
                    flex.bool(len(reflections), True), reflections.flags.strong
                )
                reflections.set_flags(
                    flex.bool(len(reflections), False), reflections.flags.integrated
                )
                reflections.set_flags(
                    flex.bool(len(reflections), False),
                    reflections.flags.centroid_outlier,
                )

            outlier_sel = reflections.get_flags(reflections.flags.centroid_outlier)

            if self.settings.outlier_display == "outliers":
                reflections = reflections.select(outlier_sel)
            if self.settings.outlier_display == "inliers":
                reflections = reflections.select(~outlier_sel)

            indexed_sel = reflections.get_flags(reflections.flags.indexed)
            strong_sel = reflections.get_flags(reflections.flags.strong)
            integrated_sel = reflections.get_flags(reflections.flags.integrated)

            if self.settings.display == "indexed":
                reflections = reflections.select(indexed_sel)
            elif self.settings.display == "unindexed":
                reflections = reflections.select(strong_sel & ~indexed_sel)
            elif self.settings.display == "integrated":
                reflections = reflections.select(integrated_sel)

            if self.settings.experiment_ids:
                sel = flex.bool(len(reflections), False)
                for i in self.settings.experiment_ids:
                    sel.set_selected(reflections["id"] == i, True)
                reflections = reflections.select(sel)

        d_spacings = 1 / reflections["rlp"].norms()

        # 155 handle data from predictions *only* if that is what we have
        if "xyzobs.px.value" in self.reflections_input:
            use_column = "xyzobs.px.value"
        else:
            use_column = "xyzcal.px"

        if self.settings.d_min is not None:
            reflections = reflections.select(d_spacings >= self.settings.d_min)
        else:
            self.settings.d_min = flex.min(d_spacings)
        if self.settings.z_min is not None:
            z = reflections[use_column].parts()[2]
            reflections = reflections.select(z >= self.settings.z_min)
        else:
            z = reflections[use_column].parts()[2]
            self.settings.z_min = flex.min(z)
        if self.settings.z_max is not None:
            z = reflections[use_column].parts()[2]
            reflections = reflections.select(z <= self.settings.z_max)
        else:
            z = reflections[use_column].parts()[2]
            self.settings.z_max = flex.max(z)

        if "n_signal" in reflections:
            if self.settings.n_min is not None:
                _ns = reflections["n_signal"]
                reflections = reflections.select(_ns >= self.settings.n_min)
            else:
                _ns = reflections["n_signal"]
                self.settings.n_min = int(flex.min(_ns))

            if self.settings.n_max is not None:
                _ns = reflections["n_signal"]
                reflections = reflections.select(_ns <= self.settings.n_max)
            else:
                _ns = reflections["n_signal"]
                self.settings.n_max = int(flex.max(_ns))

        if "partiality" in reflections:
            p = reflections["partiality"]
            if self.settings.partiality_min is not None:
                sel = p >= self.settings.partiality_min
                reflections = reflections.select(sel)
                p = p.select(sel)
            else:
                self.settings.partiality_min = flex.min(p)
            if self.settings.partiality_max is not None:
                reflections = reflections.select(p <= self.settings.partiality_max)
            else:
                self.settings.partiality_max = flex.max(p)
        points = reflections["rlp"] * 100
        self.viewer.set_points(points)
        self.viewer.set_points_data(reflections)
        colors = flex.vec3_double(len(points), (1, 1, 1))

        if len(points):
            # suggested colorblind color palette
            # sorry if you have > 8 lattices!
            palette = flex.vec3_double(
                (
                    (255, 255, 255),
                    (230, 159, 0),
                    (86, 180, 233),
                    (0, 158, 115),
                    (240, 228, 66),
                    (0, 114, 178),
                    (213, 94, 0),
                    (204, 121, 167),
                )
            )
            palette *= 1 / 255
            if not self.settings.black_background:
                bkg = flex.vec3_double()
                for j in range(len(palette)):
                    bkg.append((1, 1, 1))
                palette = bkg - palette
            self.viewer.set_palette(palette)
            n = palette.size() - 1
            if (
                self.reflections.get_flags(self.reflections.flags.indexed).count(True)
                == 0
            ):
                if "imageset_id" in reflections:
                    imageset_id = reflections["imageset_id"]
                else:
                    imageset_id = reflections["id"]
                for i in range(0, flex.max(imageset_id) + 1):
                    colors.set_selected(imageset_id == i, palette[(i % n) + 1])
            else:
                colors.set_selected(reflections["id"] == -1, palette[0])
                for i in range(0, flex.max(reflections["id"]) + 1):
                    colors.set_selected(reflections["id"] == i, palette[(i % n) + 1])
        self.viewer.set_colors(colors)

    def set_beam_centre(self, panel, beam_centre):
        detector = self.experiments[0].detector
        beam = self.experiments[0].beam
        panel = int(panel)

        try:
            p = detector[panel]
        except RuntimeError:
            raise ValueError(f"Detector does not have panel index {panel}")

        # Check if the new beam centre can be set
        beam_f, beam_s = beam_centre
        trial_s0_direction = p.get_lab_coord((beam_f, beam_s))
        try:
            panel_id, beam_centre = detector.get_ray_intersection(trial_s0_direction)
        except RuntimeError:
            # beam centre calculation fails if the beam falls between panels
            raise ValueError("No detector intersection")

        if panel_id != panel:
            raise ValueError(f"Beam centre cannot be set with panel {panel}")

        # If we get this far, then the beam centre is valid for this panel
        beam.set_unit_s0(trial_s0_direction)
